#! /bin/sh

################################################################################
#                                                                              #
# Copyright (C) 2003 Andrea Biardi                                             #
#                                                                              #
# This program is free software; you can redistribute it and/or modify it      #
# under the terms of the GNU General Public License as published by the Free   #
# Software Foundation; either version 2 of the License, or (at your option)    #
# any later version.                                                           #
#                                                                              #
# This program is distributed in the hope that it will be useful, but WITHOUT  #
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or        #
# FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for     #
# more details.                                                                #
#                                                                              #
################################################################################

HELP="\
$0 controls package version information.

Usage: $0 [options] [file]

Where \"file\" defaults to \"version.ac\".

Available options:
  -h, --help        show this help text
  -n, --name        print package name
  -a, --tar-name    print package tar name
  -u, --bug-report  print e-mail address to use for bug reports
  -t, --timestamp   print version timestamp
  -v, --version     print version string
  -m, --major       print major version number
  -i, --minor       print minor version number
  -b, --build       print build number
  -M, --MAJOR       increment major version number (reset minor and build) (*)
  -I, --MINOR       increment minor version number (reset build) (*)
  -B, --BUILD       increment build number (*)
  -T, --TIMESTAMP   update timestamp (-M, -I and -B already do) (*)

Options marked with (*) cause a rewrite of \"version.ac\".

Examples:
  $0 --version
  $0 --version --BUILD --version"

# parse arguments (1st pass)

VERSION_AC=
HAVE_OPTIONS=
for OPTION
do
  case $OPTION in
    -h|--help)
      echo "$HELP"
      exit 1
      ;;
    -*)
      HAVE_OPTIONS=yes
      ;;
    *)
      if test x"$VERSION_AC" = x; then
        VERSION_AC=$OPTION
      else
        echo "Bad option: \"$OPTION\"; try \"$0 --help\"." 1>&2
        exit 1
      fi
      ;;
  esac
done

if test x"$HAVE_OPTIONS" = x; then
  echo "$HELP" 1>&2
  exit 1
fi

test x"$VERSION_AC" = x && VERSION_AC=version.ac

# quick and dirty way..

eval $(
m4 -P << EOF
m4_divert(-1)
m4_changequote([,])
`cat "$VERSION_AC"`
m4_divert(0)
NAME=m4_ifdef([LT_AC_PKG_NAME],["LT_AC_PKG_NAME"])
TARNAME=m4_ifdef([LT_AC_PKG_TARNAME],["LT_AC_PKG_TARNAME"])
BUGREPORT=m4_ifdef([LT_AC_PKG_BUGREPORT],["LT_AC_PKG_BUGREPORT"])
TIMESTAMP=m4_ifdef([LT_AC_PKG_TIMESTAMP],["LT_AC_PKG_TIMESTAMP"])
V_MAJOR=m4_ifdef([LT_AC_PKG_V_MAJOR],["LT_AC_PKG_V_MAJOR"])
V_MINOR=m4_ifdef([LT_AC_PKG_V_MINOR],["LT_AC_PKG_V_MINOR"])
V_BUILD=m4_ifdef([LT_AC_PKG_V_BUILD],["LT_AC_PKG_V_BUILD"])
EOF
)

# keep track of changes

CHANGED=

# parse arguments (2nd pass)

for OPTION
do
  case $OPTION in
    -n|--name)
      echo "$NAME"
      ;;
    -a|--tar-name)
      echo "$TARNAME"
      ;;
    -u|--bug-report)
      echo "$BUGREPORT"
      ;;
    -t|--timestamp)
      echo "$TIMESTAMP"
      ;;
    -v|--version)
      echo "$V_MAJOR.$V_MINOR.$V_BUILD"
      ;;
    -m|--major)
      echo "$V_MAJOR"
      ;;
    -i|--minor)
      echo "$V_MINOR"
      ;;
    -b|--build)
      echo "$V_BUILD"
      ;;
    -T|--TIMESTAMP)
      TIMESTAMP=`date -u`
      CHANGED=yes
      ;;
    -M|--MAJOR)
      TIMESTAMP=`date -u`
      V_MAJOR=`expr $V_MAJOR + 1`
      V_MINOR=0
      V_BUILD=0
      CHANGED=yes
      ;;
    -I|--MINOR)
      TIMESTAMP=`date -u`
      V_MINOR=`expr $V_MINOR + 1`
      V_BUILD=0
      CHANGED=yes
      ;;
    -B|--BUILD)
      TIMESTAMP=`date -u`
      V_BUILD=`expr $V_BUILD + 1`
      CHANGED=yes
      ;;
    *)
      if test x"$OPTION" = x"$VERSION_AC"; then
        :
      else
        echo "Bad option: \"$OPTION\"; try \"$0 --help\"." 1>&2
        exit 1
      fi
      ;;
  esac
done

# rewrite "version.ac" if needed

test x"$CHANGED" = x && exit

cat > "$VERSION_AC" << EOF
################################################################################
#                                                                              #
# WARNING                                                                      #
#                                                                              #
# Please don't edit this file by hand. It may have adverse effects.            #
#                                                                              #
# This file is maintained by "version.sh" -- if you know what you are doing    #
# and still want to make your own modifications, at least have a look at the   #
# script first.                                                                #
#                                                                              #
################################################################################

m4_define([LT_AC_PKG_NAME],[$NAME])
m4_define([LT_AC_PKG_TARNAME],[$TARNAME])
m4_define([LT_AC_PKG_BUGREPORT],[$BUGREPORT])
m4_define([LT_AC_PKG_TIMESTAMP],[$TIMESTAMP])
m4_define([LT_AC_PKG_V_MAJOR],[$V_MAJOR])
m4_define([LT_AC_PKG_V_MINOR],[$V_MINOR])
m4_define([LT_AC_PKG_V_BUILD],[$V_BUILD])
m4_define([LT_AC_PKG_VERSION],[LT_AC_PKG_V_MAJOR.LT_AC_PKG_V_MINOR.LT_AC_PKG_V_BUILD])
EOF
