/*******************************************************************************

Copyright (C) 2003 Andrea Biardi - LOGICOMM

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

*******************************************************************************/

#ifndef __IBXX_ARRAY_H__
#define __IBXX_ARRAY_H__

/*******************************************************************************

Array class and functions to read/write array columns.

Compared to the Interbase API, the interface presented here is somehow
simplified:
  - Multiple dimensions are normalized to a single unit (the total number of
    items is preserved, of course); also, data is always retrieved and stored in
    row-major order.
  - Reading and writing operations are performed on the entire content of the
    array; it's not possible to read (or write) a slice of it.

In addition, though arrays seem to accept most data types available for regular
columns, some combinations may not be supported; the documentation itself
("Interbase 6 API Guide", page 158) is a bit vague on this topic; as a result,
this code does not implement the following types:

  blr_quad      Documented (page 264); don't know how to create such a column.
  blr_int64     Not documented (found in ibase.h); in dialect 3, such a column
                can be created with numeric(18,2) for example; apparently it's
                not supported by isc_array_put_slice().
  blr_d_float   Not documented (found in ibase.h); don't know how to create such
                a column.
  blr_text2     Documented (page 264); don't know how to create such a column.
  blr_varying2  Documented (page 265); don't know how to create such a column.
  blr_cstring   Documented (page 265); don't know how to create such a column.
  blr_cstring2  Documented (page 265); don't know how to create such a column.
  blr_blob      Not documented (found in ibase.h); don't know how to create such
                a column.
  blr_blob_id   Documented (page 265); don't know how to create such a column.

*******************************************************************************/

#include <cassert>
#include <cstring>
#include <ctime>
#include <iostream>
#include <libinterbasexx/config.h>
#include <libinterbasexx/cxxbits/cbuffer.h>
#include <libinterbasexx/cxxbits/format.h>
#include <libinterbasexx/cxxbits/strmisc.h>
#include <libinterbasexx/cxxbits/timestamp.h>
#include <libinterbasexx/database.h>
#include <libinterbasexx/error.h>
#include <libinterbasexx/numeric.h>
#include <libinterbasexx/transaction.h>
#include <string>
#include <ibase.h>

namespace IB {

class IBXX_EXPORT Array {
  private:
    ISC_QUAD fid; // array ID
  public:
    Array() { std::memset(&fid,0,sizeof fid); }
    Array(const ISC_QUAD& id): fid(id) {}
    const ISC_QUAD& id() const { return fid; }
    std::ostream& put(std::ostream& stream) const;
};

inline std::ostream& operator<<(std::ostream& stream, const Array& v) {
  return v.put(stream);
}

class IBXX_EXPORT ArrayDesc {
  private:
    ISC_ARRAY_DESC fdesc; // array descriptor
    int fsize;            // number of items in the array
  public:
    ArrayDesc(const Database& database, const Transaction& transaction, const std::string& table, const std::string& column);
    const ISC_ARRAY_DESC* desc() const { return &fdesc; }
    std::string table() const { return cstrncpy_c(fdesc.array_desc_relation_name,sizeof fdesc.array_desc_relation_name); }
    std::string column() const { return cstrncpy_c(fdesc.array_desc_field_name,sizeof fdesc.array_desc_field_name); }
    int size() const { return fsize; }
};

inline EInterbase cannotConvertFromArrayItem(const ArrayDesc& arrayDesc, const std::string& toType) {
  return EInterbase(format("%s: Cannot convert array item to %s (data type = %d)",arrayDesc.column(),toType,arrayDesc.desc()->array_desc_dtype));
}

inline EInterbase cannotConvertToArrayItem(const ArrayDesc& arrayDesc, const std::string& fromType) {
  return EInterbase(format("%s: Cannot convert %s to array item (data type = %d)",arrayDesc.column(),fromType,arrayDesc.desc()->array_desc_dtype));
}

inline ISC_LONG arrayBufferSize(const ArrayDesc& arrayDesc) {

  // returns the required buffer size for the given array descriptor, which is
  // normally the number of items times the item size
  //
  // for blr_varying, however, sizeof(short) is not included in what
  // array_desc_length reports and must be taken into account separately

  ISC_LONG z = arrayDesc.desc()->array_desc_length;
  if (arrayDesc.desc()->array_desc_dtype == blr_varying) z += sizeof(short);

  return z*arrayDesc.size();

}

ISC_LONG IBXX_EXPORT readArrayBuffer(const Database& database, const Transaction& transaction, const ArrayDesc& arrayDesc, const Array& array, void* data, ISC_LONG size);
ISC_LONG IBXX_EXPORT readArrayBuffer(const Database& database, const Transaction& transaction, const ArrayDesc& arrayDesc, const Array& array, CBuffer& buffer);

Array IBXX_EXPORT writeArrayBuffer(const Database& database, const Transaction& transaction, const ArrayDesc& arrayDesc, const void* data, ISC_LONG size);

inline Array writeArrayBuffer(const Database& database, const Transaction& transaction, const ArrayDesc& arrayDesc, const CBuffer& buffer) {
  return writeArrayBuffer(database,transaction,arrayDesc,buffer.data(),buffer.size());
}

#define _ARRAYITEMPTR(type,data,index) (((type*)(data))+index)
#define _ARRAYITEMSTR(size,data,index) (((char*)(data))+(index)*(size))

template<class T> T arrayItem_n(const ArrayDesc& arrayDesc, const void* data, int index) {

  // the T class should provide the following constructors:
  //   - T(short)
  //   - T(long)
  //   - T(float)
  //   - T(double)
  //   - T(const T&)
  //
  // it should also provide the /= operator used to unscale numbers, since
  // blr_short and blr_long might be scaled floating point values

  switch (arrayDesc.desc()->array_desc_dtype) {

    case blr_short: {
      IBXX_ASSERT(arrayDesc.desc()->array_desc_length == sizeof(short));
      T result = *_ARRAYITEMPTR(short,data,index);
      for (int i = arrayDesc.desc()->array_desc_scale; i < 0; ++i) result /= 10;
      return result;
    }

    case blr_long: {
      IBXX_ASSERT(arrayDesc.desc()->array_desc_length == sizeof(long));
      T result = *_ARRAYITEMPTR(long,data,index);
      for (int i = arrayDesc.desc()->array_desc_scale; i < 0; ++i) result /= 10;
      return result;
    }

    case blr_float:
      IBXX_ASSERT(arrayDesc.desc()->array_desc_length == sizeof(float));
      return T(*_ARRAYITEMPTR(float,data,index));

    case blr_double:
      IBXX_ASSERT(arrayDesc.desc()->array_desc_length == sizeof(double));
      return T(*_ARRAYITEMPTR(double,data,index));

    default:
      throw cannotConvertFromArrayItem(arrayDesc,"number");

  }

}

template<class T> T arrayItem_t(const ArrayDesc& arrayDesc, const void* data, int index) {

  // the T class should provide the following constructors:
  //   - T(const std::tm&)
  //   - T(const T&)

  switch (arrayDesc.desc()->array_desc_dtype) {

    case blr_timestamp: {
      IBXX_ASSERT(arrayDesc.desc()->array_desc_length == sizeof(ISC_TIMESTAMP));
      std::tm result;
      isc_decode_timestamp(_ARRAYITEMPTR(ISC_TIMESTAMP,data,index),&result);
      return T(result);
    }

    case blr_sql_date: {
      IBXX_ASSERT(arrayDesc.desc()->array_desc_length == sizeof(ISC_DATE));
      std::tm result;
      isc_decode_sql_date(_ARRAYITEMPTR(ISC_DATE,data,index),&result);
      IBXX_ASSERT(!result.tm_hour);
      IBXX_ASSERT(!result.tm_min);
      IBXX_ASSERT(!result.tm_sec);
      IBXX_ASSERT(!result.tm_isdst);
      return T(result);
    }

    case blr_sql_time: {
      IBXX_ASSERT(arrayDesc.desc()->array_desc_length == sizeof(ISC_TIME));
      std::tm result;
      isc_decode_sql_time(_ARRAYITEMPTR(ISC_TIME,data,index),&result);
      IBXX_ASSERT(!result.tm_year);
      IBXX_ASSERT(!result.tm_mon);
      IBXX_ASSERT(!result.tm_mday);
      IBXX_ASSERT(!result.tm_wday);
      IBXX_ASSERT(!result.tm_yday);
      return T(result);
    }

    default:
      throw cannotConvertFromArrayItem(arrayDesc,"timestamp");

  }

}

template<class T> T arrayItem_s(const ArrayDesc& arrayDesc, const void* data, int index) {

  // the T class should provide the following constructors:
  //   - T(const std::string&)
  //   - T(const T&)

  switch (arrayDesc.desc()->array_desc_dtype) {

    case blr_text: {
      char* p = _ARRAYITEMSTR(arrayDesc.desc()->array_desc_length,data,index);
      std::string result;
      return T(result.assign(p,arrayDesc.desc()->array_desc_length));
    }

    // blr_varying is a bit funny: a short is supposed to hold the actual string
    // length (as for regular VARCHAR columns), but it is seems to be always
    // zero; in addition, it comes *after* the string, so the string is actually
    // always null-terminated

    case blr_varying: {
      char* p = _ARRAYITEMSTR(arrayDesc.desc()->array_desc_length+sizeof(short),data,index);
      IBXX_ASSERT(!*(short*)(p+arrayDesc.desc()->array_desc_length));
      std::string result(p);
      return T(result);
    }

    default:
      throw cannotConvertFromArrayItem(arrayDesc,"string");

  }

}

inline int arrayItem_i(const ArrayDesc& arrayDesc, const void* data, int index) {
  return arrayItem_n<int>(arrayDesc,data,index);
}

inline double arrayItem_f(const ArrayDesc& arrayDesc, const void* data, int index) {
  return arrayItem_n<double>(arrayDesc,data,index);
}

inline Numeric arrayItem_n(const ArrayDesc& arrayDesc, const void* data, int index) {
  return arrayItem_n<Numeric>(arrayDesc,data,index);
}

inline Timestamp arrayItem_t(const ArrayDesc& arrayDesc, const void* data, int index) {
  return arrayItem_t<Timestamp>(arrayDesc,data,index);
}

inline std::string arrayItem_s(const ArrayDesc& arrayDesc, const void* data, int index) {
  return arrayItem_s<std::string>(arrayDesc,data,index);
}

template<class T> void setArrayItem_n(const ArrayDesc& arrayDesc, void* data, int index, const T& value) {

  // the T class should provide the following constructors:
  //   - T(const T&)
  //
  // in addition, it must be possible to use T to initialize the following
  // types:
  //   - float
  //   - double
  //
  // finally, T should also provide the *= operator used to scale numbers,
  // since blr_short and blr_long might be scaled floating point values

  switch (arrayDesc.desc()->array_desc_dtype) {

    case blr_short: {
      IBXX_ASSERT(arrayDesc.desc()->array_desc_length == sizeof(short));
      T t(value);
      for (int i = arrayDesc.desc()->array_desc_scale; i < 0; ++i) t *= 10;
      *_ARRAYITEMPTR(short,data,index) = short(t);
      return;
    }

    case blr_long: {
      IBXX_ASSERT(arrayDesc.desc()->array_desc_length == sizeof(long));
      T t(value);
      for (int i = arrayDesc.desc()->array_desc_scale; i < 0; ++i) t *= 10;
      *_ARRAYITEMPTR(long,data,index) = long(t);
      return;
    }

    case blr_float:
      IBXX_ASSERT(arrayDesc.desc()->array_desc_length == sizeof(float));
      *_ARRAYITEMPTR(float,data,index) = float(value);
      return;

    case blr_double:
      IBXX_ASSERT(arrayDesc.desc()->array_desc_length == sizeof(double));
      *_ARRAYITEMPTR(double,data,index) = double(value);
      return;

    default:
      throw cannotConvertToArrayItem(arrayDesc,"number");

  }

}

template<class T> void setArrayItem_t(const ArrayDesc& arrayDesc, void* data, int index, const T& value) {

  // the T class should provide the following constructors:
  //   - T(const std::tm&)
  //   - T(const T&)

  switch (arrayDesc.desc()->array_desc_dtype) {

    case blr_timestamp: {
      IBXX_ASSERT(arrayDesc.desc()->array_desc_length == sizeof(ISC_TIMESTAMP));
      std::tm m = Timestamp(value);
      return isc_encode_timestamp(&m,_ARRAYITEMPTR(ISC_TIMESTAMP,data,index));
    }

    case blr_sql_date: {
      IBXX_ASSERT(arrayDesc.desc()->array_desc_length == sizeof(ISC_DATE));
      std::tm m = Timestamp(value);
      return isc_encode_sql_date(&m,_ARRAYITEMPTR(ISC_DATE,data,index));
    }

    case blr_sql_time: {
      IBXX_ASSERT(arrayDesc.desc()->array_desc_length == sizeof(ISC_TIME));
      std::tm m = Timestamp(value);
      return isc_encode_sql_time(&m,_ARRAYITEMPTR(ISC_TIME,data,index));
    }

    default:
      throw cannotConvertToArrayItem(arrayDesc,"timestamp");

  }

}

template<class T> void setArrayItem_s(const ArrayDesc& arrayDesc, void* data, int index, const T& value) {

  // the T class should provide the following constructors:
  //   - T(const std::string&)
  //   - T(const T&)

  switch (arrayDesc.desc()->array_desc_dtype) {

    case blr_text: {
      char* p = _ARRAYITEMSTR(arrayDesc.desc()->array_desc_length,data,index);
      std::memset(p,' ',arrayDesc.desc()->array_desc_length);
      std::string(value).copy(p,arrayDesc.desc()->array_desc_length);
      return;
    }

    // for blr_varying, see comments in arrayItem_s()

    case blr_varying: {
      char* p = _ARRAYITEMSTR(arrayDesc.desc()->array_desc_length+sizeof(short),data,index);
      std::memset(p,0,arrayDesc.desc()->array_desc_length+sizeof(short));
      std::string(value).copy(p,arrayDesc.desc()->array_desc_length);
      return;
    }

    default:
      throw cannotConvertToArrayItem(arrayDesc,"string");

  }

}

} // namespace IB

#endif // __IBXX_ARRAY_H__
