/*******************************************************************************

Copyright (C) 2003 Andrea Biardi - LOGICOMM

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

*******************************************************************************/

#include <algorithm>
#include <libinterbasexx/blob.h>
#include <libinterbasexx/error.h>

namespace IB {

std::ostream& Blob::put(std::ostream& stream) const {
  return stream << "<blob:" << fid.isc_quad_high << ':' << fid.isc_quad_low << '>';
}

ISC_LONG readBlob(const Database& database, const Transaction& transaction, const Blob& blob, void* data, ISC_LONG maxsize) {

  // no blob? return 0

  if (!blob.id().isc_quad_high && !blob.id().isc_quad_low) return 0;

  // initialize things

  ISC_STATUS v[20];

  isc_db_handle dh = database.handle();
  isc_tr_handle th = transaction.handle();

  isc_blob_handle bh = 0;
  ISC_QUAD id = blob.id();

  // open the blob (this provides the handle); from this moment on, watch out
  // for exceptions, since we need to close the blob before leaving
  //
  // we do not want, however, to use a try..catch construct, which may have a
  // considerable impact on performance and makes coding more difficult

  if (isc_open_blob2(v,&dh,&th,&bh,&id,0,0)) throw EInterbase(v);

  // now ask for its size

  char items = isc_info_blob_total_length;
  char buf[32];

  if (isc_blob_info(v,&bh,1,&items,sizeof buf,buf)) {
    EInterbase e(v);
    isc_close_blob(v,&bh);
    throw e;
  }

  ISC_LONG result = (buf[0] == isc_info_blob_total_length) ? isc_vax_integer(buf+3,short(isc_vax_integer(buf+1,2))) : 0;
  if (maxsize && result > maxsize) result = maxsize;

  // if we have a buffer, read and store up to this number of bytes there,
  // calling isc_get_segment() repeatedly and asking for as much data as we can
  // read in a single pass; throw an exception if anything goes wrong

  if (char* p = (char*) data) {
    unsigned short n;
    for (ISC_LONG z = result; z > 0; z -= n, p += n) {
      if (isc_get_segment(v,(isc_blob_handle*) &bh,&n,short(std::min(z,ISC_LONG(USHRT_MAX))),p) && v[1] != isc_segment) {
        EInterbase e(v);
        isc_close_blob(v,&bh);
        throw e;
      }
    }
  }

  // don't forget to close the blob before leaving

  if (isc_close_blob(v,&bh)) throw EInterbase(v);
  return result;

}

ISC_LONG readBlob(const Database& database, const Transaction& transaction, const Blob& blob, CBuffer& buffer, ISC_LONG maxsize) {

  // get the blob size, resize the buffer, and read in the actual content

  ISC_LONG z = blobSize(database,transaction,blob,maxsize);
  if (z > buffer.size()) buffer.resize(IBXX_ALIGN_SIZE(z));

  return readBlob(database,transaction,blob,buffer.data(),buffer.size());

}

std::string readBlob(const Database& database, const Transaction& transaction, const Blob& blob, ISC_LONG maxsize) {

  // if we are given a size, trust the programmer and use it right away to
  // allocate a buffer big enough; most likely, he/she is trying to get the
  // initial part of a (text) blob
  //
  // note that the two code paths invoke two slightly different versions of
  // CBuffer::CBuffer() and readBlob()

  std::string result;

  if (maxsize) {
    CBuffer b(maxsize);
    if (ISC_LONG n = readBlob(database,transaction,blob,b.data(),maxsize)) result.assign((const char*) b.data(),n);
  } else {
    CBuffer b;
    if (readBlob(database,transaction,blob,b,maxsize)) result.assign((const char*) b.data(),b.size());
  }

  return result;

}

Blob writeBlob(const Database& database, const Transaction& transaction, const void* data, ISC_LONG size) {

  // initialize things

  ISC_STATUS v[20];
  ISC_QUAD result;

  isc_db_handle dh = database.handle();
  isc_tr_handle th = transaction.handle();

  isc_blob_handle bh = 0;

  // create a new blob (this provides both the handle and the new ID); from this
  // moment on, watch out for exceptions, since we need to either close or
  // cancel the blob before leaving
  //
  // as in readBlob(), we're not using a try..catch construct

  if (isc_create_blob2(v,&dh,&th,&bh,&result,0,0)) throw EInterbase(v);

  // call isc_put_segment() as many times as needed; jump out on error

  char* p;
  unsigned short n;

  for (p = (char*) data; size; size -= n, p += n) {
    n = short(std::min(size,ISC_LONG(USHRT_MAX)));
    if (isc_put_segment(v,&bh,n,p)) {
      EInterbase e(v);
      isc_cancel_blob(v,&bh);
      throw e;
    }
  }

  // close the blob; if an error occurs, probably we won't be able to cancel the
  // blob, but it's worth trying anyway..

  if (isc_close_blob(v,&bh)) {
    EInterbase e(v);
    isc_cancel_blob(v,&bh);
    throw e;
  }

  return Blob(result);

}

} // namespace IB
