/*******************************************************************************

Copyright (C) 2005 Andrea Biardi

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

*******************************************************************************/

#include <algorithm>
#include <cassert>
#include <cstdio>
#include <libinterbasexx/cxxbits/format.h>

namespace IB {

Format::Format(const std::string& fmt): ffmt(fmt), fnargs(0) {

  // parse the format string, construct ffields

  ffields.reserve(16);

  int n = 0;
  std::string::size_type i = 0;

  while (i < ffmt.length()) {

    Field f;
    std::string::size_type j;

    // look for the next '%'

    if ((j = ffmt.find('%',i)) == std::string::npos) {
      f.i0 = i;
      f.i1 = ffmt.length();
      f.index = 0;
      ffields.push_back(f);
      break;
    }

    // literal

    if (j != i) {
      f.i0 = i;
      f.i1 = j;
      f.index = 0;
      ffields.push_back(f);
      i = j;
    }

    // ffmt[i] == '%'

    IBXX_ASSERT(i < ffmt.length());
    IBXX_ASSERT(ffmt[i] == '%');

    if ((j = i+1) >= ffmt.length()) throw EConv("Unexpected end of string",ffmt,j);

    // but maybe it's just "%%"

    if (ffmt[j] == '%') {
      f.i0 = i;
      f.i1 = j;
      f.index = 0;
      ffields.push_back(f);
      i = ++j;
      continue;
    }

    // check for an argument index (a decimal number ended by $); its sign is
    // eaten up, but the resulting index must not be negative

    f.index = 0;

    if (j < ffmt.length()) {
      std::string::size_type k = j;
      if (ffmt[k] == '-' || ffmt[k] == '+') ++k;
      while (k < ffmt.length() && ffmt[k] >= '0' && ffmt[k] <= '9') ++k;
      if (k < ffmt.length() && ffmt[k] == '$') {
        if (!a2i(ffmt.substr(j,k-j),f.index) || f.index < 0) throw EConv("Bad argument index",ffmt,j);
        j = k+1;
      }
    }

    if (!f.index) f.index = ++n;
    IBXX_ASSERT(f.index > 0);

    // anything that is not a digit, a '-' sign, a dot or a conversion specifier
    // is blindly taken as an optional flag; this leaves some room for
    // customized flags

    f.flags.clear();

    if (j < ffmt.length()) {
      std::string::size_type k;
      if ((k = ffmt.find_first_of("-0123456789.cdfiopsxXy",j)) == std::string::npos) k = ffmt.length();
      f.flags = ffmt.substr(j,k-j);
      j = k;
    }

    // width is a decimal number, possibly with a leading '-' sign; the default
    // width is zero
    //
    // note: a positive width should not come with a leading '+'; in such a
    // case, the '+' character would be taken as a flag

    f.width = 0;

    if (j < ffmt.length()) {
      std::string::size_type k = j;
      if (ffmt[k] == '-') ++k;
      while (k < ffmt.length() && ffmt[k] >= '0' && ffmt[k] <= '9') ++k;
      if (k != j) {
        if (!a2i(ffmt.substr(j,k-j),f.width)) throw EConv("Bad field width",ffmt,j);
        j = k;
      }
    }

    // precision is a dot followed by another decimal number, possibly with a
    // leading '-' sign (an explicit leading '+' is not allowed); a dot alone is
    // enough (default precision is zero)

    f.precision = 0;

    if (j < ffmt.length() && ffmt[j] == '.') {
      if (++j < ffmt.length()) {
        std::string::size_type k = j;
        if (ffmt[k] == '-') ++k;
        while (k < ffmt.length() && ffmt[k] >= '0' && ffmt[k] <= '9') ++k;
        if (k != j) {
          if (!a2i(ffmt.substr(j,k-j),f.precision)) throw EConv("Bad field precision",ffmt,j);
          j = k;
        }
      }
    }

    // we should encounter the conversion specifier now

    if (j >= ffmt.length()) throw EConv("Missing conversion specifier",ffmt,j);
    if (std::string("cdfiopsxXy").find(ffmt[j]) == std::string::npos) throw EConv("Bad conversion specifier",ffmt,j);

    f.conv = ffmt[j];

    // push_back() this field..

    f.i0 = i;
    f.i1 = ++j;
    ffields.push_back(f);

    // ..and move on

    i = j;

  }

  // uncomment the following if you need to debug the parser; you may also need
  // to #include <iostream>

  // std::cerr << "Format::Format(\"" << ffmt << "\")" << std::endl;
  // for (Fields::const_iterator p = ffields.begin(); p != ffields.end(); ++p) {
  //   std::cerr << "  \"" << ffmt.substr(p->i0,p->i1-p->i0) << "\": ";
  //   if (p->index)
  //     std::cerr << "arg index = " << p->index << ", flags = \"" << p->flags << "\", width = " << p->width << ", precision = " << p->precision << ", conv = " << p->conv << std::endl;
  //   else
  //     std::cerr << "literal" << std::endl;
  // }

}

Format::operator std::string() const {

  // concatenate all fields, making sure that we have been given an argument for
  // each; copy literals from ffmt directly

  std::string result;
  result.reserve(ffmt.length()*2);

  for (Fields::const_iterator p = ffields.begin(); p != ffields.end(); ++p) {
    if (p->index > fnargs) throw Exception(format("No argument supplied for \"%s\"",ffmt.substr(p->i0,p->i1-p->i0)));
    result += p->index ? p->str : ffmt.substr(p->i0,p->i1-p->i0);
  }

  return result;

}

void Format::stdfmt_f(std::string& str, const std::string& flags, int width, int precision, char, double value) {

  // this function supports the conversion specifier 'f'; width, precision, and
  // a couple of flags are honored, but the actual formatting is currently
  // handed over to the standard library function snprintf(); it will be
  // reimplemented better at some other future time

  std::string fmt = "%";

  if (flags.find('+') != std::string::npos)
    fmt += '+';
  else if (flags.find(' ') != std::string::npos)
    fmt += ' ';

  if (width) fmt += int2str(width);
  if (precision) (fmt += '.') += int2str(precision);

  fmt += 'f';

  char buf[32];
  int n = std::snprintf(buf,sizeof buf,fmt.c_str(),value);

  if (n < 0 || n >= (int)sizeof buf) throw Exception("Number too big");
  str.assign(buf,n);

}

void Format::stdfmt_s(std::string& str, const std::string& flags, int width, int precision, char conv, const char* value) {

  // this function supports the following conversion specifiers:
  //
  //   conv  render value as
  //   ----  --------------------
  //   's'   string of characters
  //   'p'   pointer
  //
  // 'p' conversions are actually handed over to stdfmt_i() for hexadecimal
  // representation, so you should have a look at that function for further
  // information; what follows here applies only to 's' conversions
  //
  // the width determines the minimum field width; if the string resulting from
  // the conversion is shorter, it is padded with spaces (on the left if the
  // width is positive, on the right if the width is negative)
  //
  // the precision controls the maximum number of characters that should appear:
  // the field is truncated if it is longer (in this case, the source string
  // need not be null-terminated)
  //
  // the flags below are honored, in the following order:
  //
  //   flag  meaning
  //   ----  -------------------------------------------------------------------
  //   'U'   the string is converted to uppercase
  //   'L'   the string is converted to lowercase
  //   '/'   a path delimiter is appended to the string, if needed; see pdelim()
  //   '"'   the string is quoted; see strquote()
  //
  // warning: the '/" and '"' flags often make the resulting string longer, but
  // since they are applied after the string is constructed, the maximum number
  // of characters (precision) may not be honored: these flags are not meant to
  // be used with a precision, which may make them partially or completely
  // useless anyway (think '/' and pathname concatenation)

  if (conv == 'p') {
    IBXX_ASSERT(sizeof(unsigned int) == sizeof(void*));
    return stdfmt_i(str,"",width,std::max(precision,2*(int)sizeof(void*)),'x',(unsigned int)value);
  }

  if (precision > 0)
    cstrncpy(str,value,precision);
  else
    str = value;

  if (flags.find('U') != std::string::npos)
    ucase(str);
  else if (flags.find('L') != std::string::npos)
    lcase(str);

  if (flags.find('/') != std::string::npos) pdelim(str);

  if (flags.find('"') != std::string::npos) strquote(str);

  width >= 0 ? lpad(str,width) : rpad(str,-width);

}

} // namespace IB
