/*******************************************************************************

Copyright (C) 2005 Andrea Biardi

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

*******************************************************************************/

#ifndef __IBXX_CXXBITS_FORMAT_H__
#define __IBXX_CXXBITS_FORMAT_H__

/*******************************************************************************

This code implements a printf()-like function, which renders an arbitrary number
of arguments according to a given format string, and returns the formatted
string; in this respect, since it does not produce any output on its own, it is
more similar to sprintf() than to printf().

The basic idea is that the Format class has just the ability to parse the format
string, and then relies on specialized helpers for the actual rendering of the
arguments, as they come in through operator<<(). This allows for some
flexibility when it comes to custom data types (see below).

You can use the Format class directly, or call one of the overloaded convenience
functions format(), as in:

  std::string s = Format("%d %s") << 3 << "apples";    // "3 apples"
  std::string t = format("%d %s",3,"mangoes");         // "3 mangoes"

The rules for the format string are basically a subset of the ones that control
the standard printf() function, with some slight variations.

Each field has the general format:

  %[index$][flags][[-]width][.[[-]precision]]{c|d|f|i|o|p|s|x|X|y}

A literal '%' may be included in the format string by means of "%%".

First, the index allows to rearrange the arguments according to a specific order
(arguments are numbered starting from one):

  std::string s = format("%2$d %1$s","bananas",3);     // "3 bananas"

Each argument can be referenced zero or more times in the format string, and
excess arguments are silently dropped:

  std::string s = format("%2$d %s %d","pears",3,4,5);  // "3 pears 3"

Also, as you can see from the example above, any field without an explicit index
is assigned the next available number, starting from one. An explicit index
never alters nor restarts the sequence.

A sequence of zero or more flag characters may then follow. These flags are
meant to alter the default behavior of the conversion, and the actual meaning
varies with the helper being used (which, in turn, depends on the data type of
the argument).

In the following example, the flag 'U' turns the string to uppercase:

  std::string s = format("%d %Us",3,"coconuts");       // "3 COCONUTS"

As a general rule, you can expect unrecognized flags to be ignored, but this
depends on the helper.

Width and precision control the amount of space used up by the argument, and
behave exactly as in printf():

  std::string s = format("%.4d %10s",3,"oranges");     // "0003    oranges"

Finally, the conversion specifier determines the actual representation of the
argument:

  std::string s = format("%c %d %o %x",65,65,65,65);   // "A 65 101 41"

Standard helpers are provided for bool, signed and unsigned char, short, int and
long, as well as for IBXX_INT64 and CXXBITS_UINT64 if defined. These helpers
support the 'c', 'd', 'i', 'o', 'x', 'X' and 'y' conversion specifiers.

For float and double, use the conversion specifier 'f'.

'p' may be used with any pointer type (though a cast to [const] void* is
probably needed), while [const] char* and std::string use 's'.

Available conversion specifiers and flags are described in stdfmt_i(),
stdfmt_s() and friends.

As said above, one of the major advantages of format(), compared to printf(), is
the ability of rendering custom data types:

  struct Point {
    int x;
    int y;
  };

  template<> struct FormatHelper<Point> {
    static void fmt(...) { ... }
  };

For custom data types, it is recommended to use the 's' conversion specifier and
then provide specific conversion flags, if need be.

*******************************************************************************/

#include <libinterbasexx/cxxbits/config.h>
#include <libinterbasexx/cxxbits/exception.h>
#include <libinterbasexx/cxxbits/strmisc.h>
#include <string>
#include <vector>

namespace IB {

// forward declarations

template<class T0> std::string format(const std::string& fmt, T0 arg0);
template<class T0, class T1> std::string format(const std::string& fmt, T0 arg0, T1 arg1);
template<class T0, class T1, class T2> std::string format(const std::string& fmt, T0 arg0, T1 arg1, T2 arg2);
template<class T0, class T1, class T2, class T3> std::string format(const std::string& fmt, T0 arg0, T1 arg1, T2 arg2, T3 arg3);
template<class T0, class T1, class T2, class T3, class T4> std::string format(const std::string& fmt, T0 arg0, T1 arg1, T2 arg2, T3 arg3, T4 arg4);
template<class T0, class T1, class T2, class T3, class T4, class T5> std::string format(const std::string& fmt, T0 arg0, T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5);
template<class T0, class T1, class T2, class T3, class T4, class T5, class T6> std::string format(const std::string& fmt, T0 arg0, T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5, T6 arg6);
template<class T0, class T1, class T2, class T3, class T4, class T5, class T6, class T7> std::string format(const std::string& fmt, T0 arg0, T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5, T6 arg6, T7 arg7);
template<class T0, class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8> std::string format(const std::string& fmt, T0 arg0, T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5, T6 arg6, T7 arg7, T8 arg8);
template<class T0, class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8, class T9> std::string format(const std::string& fmt, T0 arg0, T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5, T6 arg6, T7 arg7, T8 arg8, T9 arg9);

// generic helper

template<class T> struct FormatHelper {
  static void fmt(std::string&, const std::string&, int, int, char, T) { throw Exception(format("No format helper for arguments of this type: %s",typeid(T).name())); }
};

// Format class

class IBXX_EXPORT Format {
  private:
    struct Field {
      std::string::size_type i0; // index of ffmt to the first character of the field
      std::string::size_type i1; // index of ffmt to the first character after the field
      int index;                 // argument index (1-based; 0 = literal)
      std::string flags;         // format flags
      int width;                 // field width
      int precision;             // field precision
      char conv;                 // conversion specifier
      std::string str;           // formatted string
    };
    typedef std::vector<Field> Fields;
  private:
    std::string ffmt; // format string
    Fields ffields;   // broken-down (parsed) version of ffmt
    int fnargs;       // number of arguments supplied so far through operator<<()
  public:
    Format(const std::string& fmt);
    operator std::string() const;
    template<class T> Format& operator<<(T value);
    template<class T> static void stdfmt_i(std::string& str, const std::string& flags, int width, int precision, char conv, T value);
    static void stdfmt_f(std::string& str, const std::string& flags, int width, int precision, char conv, double value);
    static void stdfmt_s(std::string& str, const std::string& flags, int width, int precision, char conv, const char* value);
};

template<class T> Format& Format::operator<<(T value) {

  // increment the number of arguments supplied so far

  ++fnargs;

  // look for all fields referencing this argument, and ask the proper helper to
  // render the value
  //
  // please note that each argument may be present more than once in ffields,
  // though that isn't usually the case; on the other hand, it is equally
  // possible that the argument has been supplied, but no rendering is actually
  // required:
  //
  //   std::string s = format("%1$s %1$s","hello","world");
  //
  // the assignment above produces "hello hello", and leaves out the second
  // argument

  for (Fields::iterator p = ffields.begin(); p != ffields.end(); ++p)
    if (p->index == fnargs)
      FormatHelper<T>::fmt(p->str,p->flags,p->width,p->precision,p->conv,value);

  return *this;

}

template<class T> void Format::stdfmt_i(std::string& str, const std::string& flags, int width, int precision, char conv, T value) {

  // this function supports the following conversion specifiers:
  //
  //   conv  render value as
  //   ----  -------------------------------------------------
  //   'c'   unsigned char
  //   'd'   decimal (base 10)
  //   'i'   decimal (same as 'd')
  //   'o'   octal (base 8)
  //   'x'   hexadecimal (base 16, using lowercase characters)
  //   'X'   hexadecimal (base 16, using uppercase characters)
  //   'y'   binary (base two)
  //
  // the width determines the minimum field width: if the string resulting from
  // the conversion is shorter, it is padded with spaces (on the left if the
  // width is positive, on the right if the width is negative)
  //
  // the precision controls the minimum number of digits that should appear (a
  // missing or negative precision is treated as zero); the string resulting
  // from the conversion is left-padded with zeroes
  //
  // the following flags are honored:
  //
  //   flag  meaning
  //   ----  -------------------------------------------------------------------
  //   ' '   a blank space appears before a positive number (ignored for 'c'
  //         conversions)
  //   '+'   a sign ('+' or '-') is always placed before the number (by default,
  //         a sign is used only for negative numbers); '+' overrides ' ' if
  //         both are used (ignored for 'c' conversions)

  if (conv == 'c') {

    str = (unsigned char)(value);

  } else {

    int b;                   // base
    const char* d = LDIGITS; // lowercase or uppercase character set
    char g = 0;              // when != 0, force sign for positive numbers

    switch (conv) {

      case 'd':
      case 'i':
        b = 10;
        break;

      case 'o':
        b = 8;
        break;

      case 'X':
        d = UDIGITS; // fall-through is intentional

      case 'x':
        b = 16;
        break;

      case 'y':
        b = 2;
        break;

      default:
        throw Exception(format("Bad conversion specifier: %c",conv));

    }

    // honor the flags

    if (flags.find('+') != std::string::npos)
      g = '+';
    else if (flags.find(' ') != std::string::npos)
      g = ' ';

    // i2a() does the actual job

    str = i2a(value,precision,b,d,g);

  }

  // don't forget the width

  width >= 0 ? lpad(str,width) : rpad(str,-width);

}

// specialized helpers for standard types

template<> struct FormatHelper<bool> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, bool value) { Format::stdfmt_i(str,flags,width,precision,conv,value); }
};

template<> struct FormatHelper<char> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, char value) { Format::stdfmt_i(str,flags,width,precision,conv,value); }
};

template<> struct FormatHelper<signed char> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, signed char value) { Format::stdfmt_i(str,flags,width,precision,conv,value); }
};

template<> struct FormatHelper<unsigned char> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, unsigned char value) { Format::stdfmt_i(str,flags,width,precision,conv,value); }
};

template<> struct FormatHelper<short> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, short value) { Format::stdfmt_i(str,flags,width,precision,conv,value); }
};

template<> struct FormatHelper<unsigned short> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, unsigned short value) { Format::stdfmt_i(str,flags,width,precision,conv,value); }
};

template<> struct FormatHelper<int> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, int value) { Format::stdfmt_i(str,flags,width,precision,conv,value); }
};

template<> struct FormatHelper<unsigned int> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, unsigned int value) { Format::stdfmt_i(str,flags,width,precision,conv,value); }
};

template<> struct FormatHelper<long> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, long value) { Format::stdfmt_i(str,flags,width,precision,conv,value); }
};

template<> struct FormatHelper<unsigned long> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, unsigned long value) { Format::stdfmt_i(str,flags,width,precision,conv,value); }
};

#ifdef CXXBITS_HAVE_LONG_LONG

template<> struct FormatHelper<long long> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, long long value) { Format::stdfmt_i(str,flags,width,precision,conv,value); }
};

template<> struct FormatHelper<unsigned long long> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, unsigned long long value) { Format::stdfmt_i(str,flags,width,precision,conv,value); }
};

#endif

template<> struct FormatHelper<float> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, float value) { Format::stdfmt_f(str,flags,width,precision,conv,value); }
};

template<> struct FormatHelper<double> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, double value) { Format::stdfmt_f(str,flags,width,precision,conv,value); }
};

template<> struct FormatHelper<const void*> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, const void* value) { Format::stdfmt_s(str,flags,width,precision,conv,(const char*)value); }
};

template<> struct FormatHelper<void*> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, void* value) { Format::stdfmt_s(str,flags,width,precision,conv,(const char*)value); }
};

template<> struct FormatHelper<const char*> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, const char* value) { Format::stdfmt_s(str,flags,width,precision,conv,value); }
};

template<> struct FormatHelper<char*> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, char* value) { Format::stdfmt_s(str,flags,width,precision,conv,value); }
};

template<> struct FormatHelper<std::string> {
  static void fmt(std::string& str, const std::string& flags, int width, int precision, char conv, const std::string& value) { Format::stdfmt_s(str,flags,width,precision,conv,value.c_str()); }
};

// convenience format() functions (for one to ten arguments)

template<class T0> std::string format(const std::string& fmt, T0 arg0) { return Format(fmt) << arg0; }
template<class T0, class T1> std::string format(const std::string& fmt, T0 arg0, T1 arg1) { return Format(fmt) << arg0 << arg1; }
template<class T0, class T1, class T2> std::string format(const std::string& fmt, T0 arg0, T1 arg1, T2 arg2) { return Format(fmt) << arg0 << arg1 << arg2; }
template<class T0, class T1, class T2, class T3> std::string format(const std::string& fmt, T0 arg0, T1 arg1, T2 arg2, T3 arg3) { return Format(fmt) << arg0 << arg1 << arg2 << arg3; }
template<class T0, class T1, class T2, class T3, class T4> std::string format(const std::string& fmt, T0 arg0, T1 arg1, T2 arg2, T3 arg3, T4 arg4) { return Format(fmt) << arg0 << arg1 << arg2 << arg3 << arg4; }
template<class T0, class T1, class T2, class T3, class T4, class T5> std::string format(const std::string& fmt, T0 arg0, T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5) { return Format(fmt) << arg0 << arg1 << arg2 << arg3 << arg4 << arg5; }
template<class T0, class T1, class T2, class T3, class T4, class T5, class T6> std::string format(const std::string& fmt, T0 arg0, T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5, T6 arg6) { return Format(fmt) << arg0 << arg1 << arg2 << arg3 << arg4 << arg5 << arg6; }
template<class T0, class T1, class T2, class T3, class T4, class T5, class T6, class T7> std::string format(const std::string& fmt, T0 arg0, T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5, T6 arg6, T7 arg7) { return Format(fmt) << arg0 << arg1 << arg2 << arg3 << arg4 << arg5 << arg6 << arg7; }
template<class T0, class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8> std::string format(const std::string& fmt, T0 arg0, T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5, T6 arg6, T7 arg7, T8 arg8) { return Format(fmt) << arg0 << arg1 << arg2 << arg3 << arg4 << arg5 << arg6 << arg7 << arg8; }
template<class T0, class T1, class T2, class T3, class T4, class T5, class T6, class T7, class T8, class T9> std::string format(const std::string& fmt, T0 arg0, T1 arg1, T2 arg2, T3 arg3, T4 arg4, T5 arg5, T6 arg6, T7 arg7, T8 arg8, T9 arg9) { return Format(fmt) << arg0 << arg1 << arg2 << arg3 << arg4 << arg5 << arg6 << arg7 << arg8 << arg9; }

} // namespace IB

#endif // __IBXX_CXXBITS_FORMAT_H__
