/*******************************************************************************

Copyright (C) 2004 Andrea Biardi

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

*******************************************************************************/

#ifndef __IBXX_CXXBITS_STRMISC_H__
#define __IBXX_CXXBITS_STRMISC_H__

/*******************************************************************************

Miscellaneous string-handling functions.

C-string-related functions:

  cstrncpy()     Copies at most maxlen characters from a null-terminated C
                 string to a C++ string. Returns a reference to it.
  cstrncpy_c()   Same as cstrncpy, but returns a new string.

Case conversion:

  lcase()        Converts a string to all lowercase; returns a reference to the
                 modified string.
  lcase_c()      Same as lcase(), but returns a new string (or char).
  ucase()        Converts a string to all uppercase; returns a reference to the
                 modified string.
  ucase_c()      Same as ucase(), but returns a new string (or char).
  cap()          Capitalizes a string: the first character and all other
                 characters following a space are converted to uppercase, the
                 remaining ones are converted to lowercase; custom symbols may
                 be given as the second argument); returns a reference to the
                 modified string.
  cap_c()        Same as cap(), but returns a new string.

Trim functions:

  ltrim()        Removes one or more leading characters from a string; returns a
                 reference to the modified string.
  ltrim_c()      Same as ltrim(), but returns a new string.
  rtrim()        Removes one or more trailing characters from a string; returns
                 a reference to the modified string.
  rtrim_c()      Same as rtrim(), but returns a new string.
  trim()         Combines ltrim() and rtrim(); returns a reference to the
                 modified string.
  trim_c()       Same as trim(), but returns a new string.

Padding functions:

  lpad()         Left-fills a string up to a given number of characters (does
                 not truncate the original string); returns a reference to the
                 modified string.
  lpad_c()       Same as lpad(), but returns a new string.
  rpad()         Right-fills a string up to a given number of characters (does
                 not truncate the original string); returns a reference to the
                 modified string.
  rpad_c()       Same as rpad(), but returns a new string.
  pdelim()       Ensures a pathname ends with the trailing delimiter (a blank
                 pathname is left untouched); returns a reference to the
                 modified string. The actual delimiter is defined by
                 IBXX_PATH_DELIM.
  pdelim_c()     Same as pdelim(), but returns a new string.

Replace functions:

  strreplace()   Replaces all occurrences of a substring within a string;
                 returns a reference to the modified string.
  strreplace_c() Same as strreplace(), but returns a new string.

Quote/unquote functions:

  strquote()     Quotes a string using an arbitrary character; any preexisting
                 quote is doubled. Returns a reference to the modified string.
  strquote_c()   Same as strquote(), but returns a new string.

Conversion functions:

  a2i()          Converts a string to an integer, skipping any leading
                 whitespace; any arbitrary data type may be used. Returns false
                 if the conversion is not possible.
  str2int()      Same as a2i(), but works with int and throws an exception if
                 the conversion is not possible.
  str2intdef()   Same as str2int(), but returns a default value if the
                 conversion is not possible.
  i2a()          Converts an integer to a string; both the precision (minimum
                 number of digits) and the base may be specified.
  int2str()      Same as i2a(), but works with int.

Other functions:

  concat()       Returns a string formed by concatenating (repeating) n times a
                 given source string; an optional string will separate items. An
                 alternate version loops over all strings of an STL container
                 (for example, std::vector<std::string>).
  strbind()      Returns an STL container of std::string where each item is the
                 concatenation of corresponding items from two source containers
                 of the same type (for example, std::list<std::string>; optional
                 arguments control the width of the left column, the character
                 used to right-fill it and the string used to separate the two
                 columns.

*******************************************************************************/

#include <algorithm>
#include <cassert>
#include <cctype>
#include <cstddef>
#include <cstring>
#include <libinterbasexx/cxxbits/config.h>
#include <libinterbasexx/cxxbits/exception.h>
#include <limits>
#include <string>
#include <typeinfo>

namespace IB {

#define WHITESPACE " \t\n\r\v\f"

#define LDIGITS "0123456789abcdefghijklmnopqrstuvwxyz"
#define UDIGITS "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"

// C-string-related functions

inline std::string& cstrncpy(std::string& s, const char* p, std::size_t maxlen) {
  const char* q;
  for (q = p; *q && maxlen--; ++q);
  return s.assign(p,q-p);
}

inline std::string cstrncpy_c(const char* p, std::size_t maxlen) {
  std::string result;
  return cstrncpy(result,p,maxlen);
}

// case conversion

std::string& IBXX_EXPORT lcase(std::string& s);

inline char lcase_c(char c) {
  int result;
  return (result = std::tolower(c)) == -1 ? c : result;
}

inline std::string lcase_c(const std::string& s) {
  std::string result;
  return lcase(result = s);
}

std::string& IBXX_EXPORT ucase(std::string& s);

inline char ucase_c(char c) {
  int result;
  return (result = std::toupper(c)) == -1 ? c : result;
}

inline std::string ucase_c(const std::string& s) {
  std::string result;
  return ucase(result = s);
}

std::string& IBXX_EXPORT cap(std::string& s, const std::string& t = WHITESPACE);

inline std::string cap_c(const std::string& s, const std::string& t = WHITESPACE) {
  std::string result;
  return cap(result = s,t);
}

// trim functions

inline std::string& ltrim(std::string& s, const std::string& t = WHITESPACE) {
  return s.erase(0,s.find_first_not_of(t));
}

inline std::string ltrim_c(const std::string& s, const std::string& t = WHITESPACE) {
  std::string::size_type i;
  return (i = s.find_first_not_of(t)) == std::string::npos ? std::string("") : s.substr(i);
}

inline std::string& rtrim(std::string& s, const std::string& t = WHITESPACE) {
  return s.erase(s.find_last_not_of(t)+1);
}

inline std::string rtrim_c(const std::string& s, const std::string& t = WHITESPACE) {
  std::string::size_type i;
  return (i = s.find_last_not_of(t)) == std::string::npos ? std::string("") : s.substr(0,i+1);
}

inline std::string& trim(std::string& s, const std::string& t = WHITESPACE) {
  return ltrim(rtrim(s,t));
}

inline std::string trim_c(const std::string& s, const std::string& t = WHITESPACE) {
  std::string result;
  return trim(result = s,t);
}

// padding functions

inline std::string& lpad(std::string& s, int n, char c = ' ') {
  return s.size() < std::string::size_type(n) ? s.insert(0,std::string(n-s.size(),c)) : s;
}

inline std::string lpad_c(const std::string& s, int n, char c = ' ') {
  std::string result;
  return lpad(result = s,n,c);
}

inline std::string& rpad(std::string& s, int n, char c = ' ') {
  return s.size() < std::string::size_type(n) ? s.append(std::string(n-s.size(),c)) : s;
}

inline std::string rpad_c(const std::string& s, int n, char c = ' ') {
  std::string result;
  return rpad(result = s,n,c);
}

inline std::string& pdelim(std::string& s) {
  if (!s.empty() && s[s.length()-1] != IBXX_PATH_DELIM) s += IBXX_PATH_DELIM;
  return s;
}

inline std::string pdelim_c(const std::string& s) {
  std::string result;
  return pdelim(result = s);
}

// replace functions

std::string& IBXX_EXPORT strreplace(std::string& s, const std::string& t, const std::string& u);

inline std::string strreplace_c(const std::string& s, const std::string& t, const std::string& u) {
  std::string result;
  return strreplace(result = s,t,u);
}

// quote/unquote functions

std::string& IBXX_EXPORT strquote(std::string& s, char c = '"');

inline std::string strquote_c(const std::string& s, char c = '"') {
  std::string result;
  return strquote(result = s,c);
}

// conversion functions

template<class T> bool a2i(const char* s, T& n) {

  // attempt an ascii to integer conversion (signed, base 10); return true if
  // the conversion is successful

  // skip leading whitespace

  while (*s && std::strchr(WHITESPACE,*s)) ++s;

  // check for an optional sign

  bool e = false; // true if negative sign found

  switch (*s) {
    case '-':
      e = true; // fall-through is intentional
    case '+':
      ++s;
  };

  // if no digits, the conversion fails

  if (*s < '0' || *s > '9') return false;

  // convert one digit at a time; follow two different code paths depending on
  // the sign (this should optimize things a bit)

  n = 0;

  if (e) {
    while (char c = *s++) {
      if (c < '0' || c > '9') return false;
      n *= 10;
      n -= c-'0';
    }
  } else {
    while (char c = *s++) {
      if (c < '0' || c > '9') return false;
      n *= 10;
      n += c-'0';
    }
  }

  return true;

}

template<class T> bool a2i(const std::string& s, T& n) {
  return a2i(s.c_str(),n);
}

int IBXX_EXPORT str2int(const char* s);

inline int str2int(const std::string& s) {
  return str2int(s.c_str());
}

int IBXX_EXPORT str2intdef(const char* s, int def);

inline int str2intdef(const std::string& s, int def) {
  return str2intdef(s.c_str(),def);
}

template<class T> std::string i2a(T value, int precision = 0, int base = 10, const char* digits = LDIGITS, char psign = 0) {

  // convert an integer value to its ascii representation; both the base (by
  // default, 10) and the precision (minimum number of digits) may be specified
  //
  // digits must contain an appropriate number of characters for the base being
  // requested: the default character set allows for up to base 36, using
  // lowercase characters 'a' to 'z' in addition do decimal digits
  //
  // for base 10 representation, only negative values are rendered with a sign
  // character; in this base, psign may be used to force the sign character to
  // appear for non-negative values as well ('+' and ' ' are common choices)
  //
  // for bases other than 10, complement notation is used and no sign is ever
  // inserted

  IBXX_ASSERT(base >= 2);
  IBXX_ASSERT(std::string::size_type(base) <= std::strlen(digits));

  char s[80];                        // temporary buffer, max number of digits
  char* t = s+sizeof s;              // t = 1 char past the end of the buffer
  precision = std::max(precision,1); // minimum precision is actually 1 digit

  // this one used to be:
  //
  //   if (value >= 0) ..
  //
  // but it turned out that it generates a lot of annoying warnings (comparison
  // is always true) when T is unsigned

  if ((value > 0) || (!value)) {

    // non-negative number

    for (int i = 0; (i < precision) || (value); ++i) {
      if (t == s) throw EMath("Number too big");
      int z = value % base;
      *--t = digits[z];
      value /= base;
    }

    // for base 10 representation, optionally insert the sign character

    if ((base == 10) && (psign)) {
      if (t == s) throw EMath("Number too big");
      *--t = psign;
    }

  } else if (base == 10) {

    // negative number, base 10: always use the negative sign character '-'
    //
    // note that we don't change the sign of the value, as this may lead to
    // overflow, so we have to stick to a negative number

    for (int i = 0; (i < precision) || (value); ++i) {
      if (t == s) throw EMath("Number too big");
      int z = value % base;
      *--t = digits[-z];
      value /= base;
    }

    if (t == s) throw EMath("Number too big");
    *--t = '-';

  } else {

    // negative number, arbitrary base: use complement notation
    //
    // when negating the value, we use (T)0-value in order to suppress some
    // harmless compiler warnings when T is unsigned (this code path cannot be
    // taken anyway)

    ++value;
    value = (T)0-value;

    // then, obtain a reference value, so that we know something about the
    // minimum number of digits required by the type T; when rendering a
    // negative 32-bit integer in base 16, for example, the minimum precision
    // will be four digits

    T rfval = std::numeric_limits<T>::min();
#ifdef __BORLANDC__
    if (typeid(T) == typeid(__int64)) rfval = (__int64)((__int64)1 << 63);
#endif

    // now, the real loop

    for (int i = 0; (i < precision) || (rfval); ++i) {
      if (t == s) throw EMath("Number too big");
      int z = (base-1)-(int)(value % base);
      *--t = digits[z];
      value /= base;
      rfval /= base;
    }

    // at this point, value should be zero; this ensures that we have rendered
    // all digits

    IBXX_ASSERT(!value);

  }

  // check how many characters we have used, and return a copy of the string

  int n = s+sizeof s-t;
  IBXX_ASSERT(n >= 1);
  IBXX_ASSERT(std::string::size_type(n) <= sizeof s);

  return cstrncpy_c(t,n);

}

std::string IBXX_EXPORT int2str(int value, int precision = 0, int base = 10, const char* digits = LDIGITS, char psign = 0);

// other functions

std::string IBXX_EXPORT concat(const std::string& s, int n, const std::string& t = "");

inline std::string concat(char c, int n, const std::string& t = "") {
  return t.empty() ? std::string(n,c) : concat(std::string(1,c),n,t);
}

template<class T> std::string concat(const T& s, const std::string& t = "") {
  std::string result;
  for (typename T::const_iterator p = s.begin(); p != s.end(); ++p) {
    if (p != s.begin()) result += t;
    result += *p;
  }
  return result;
}

template<class T> T strbind(const T& l, const T& r, const std::string& t = " ", int n = -1, char u = ' ') {

  T result(std::max(l.size(),r.size()));
  typename T::const_iterator p;
  typename T::const_iterator q;

  // determine the appropriate width for the left column, if not given

  if (n < 0) for (p = l.begin(), n = 0; p != l.end(); ++p) n = std::max(n,int(p->length()));

  // arrange strings side by side

  std::string s;
  for (p = l.begin(), q = r.begin(); p != l.end() || q != r.end();) {
    s = p == l.end() ? concat(u,n) : rpad(s = *p,n,u);
    if (q != r.end()) (s += t) += *q;
    result.push_back(s);
    if (p != l.end()) ++p;
    if (q != r.end()) ++q;
  }

  return result;

}

} // namespace IB

#endif // __IBXX_CXXBITS_STRMISC_H__
