/*******************************************************************************

Copyright (C) 2005 Andrea Biardi

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

*******************************************************************************/

#ifndef __IBXX_CXXBITS_TIMESTAMP_H__
#define __IBXX_CXXBITS_TIMESTAMP_H__

/*******************************************************************************

The purpose of this class is to encapsulate a date/time value, with one-second
precision.

Values given for year, month, day, hour, minutes and seconds are interpreted
according to the current timezone (local time). str() also reports local time.

Timestamp provides overloaded constructors, assignment and arithmetic operators;
using this class should be pretty straightforward.

Keep in mind that Timestamp expresses time in seconds; in particular, if you
want to calculate the difference between two dates, use:

  Timestamp t0(2004,12,15);
  Timestamp t1(2005,6,1);
  int diff = t1-t0;

and apply some division/remainder operators to work out the number of days,
minutes or seconds.

*******************************************************************************/

#include <ctime>
#include <libinterbasexx/cxxbits/config.h>
#include <iostream>
#include <string>

namespace IB {

class IBXX_EXPORT Timestamp {
  private:
    std::time_t ftime; // seconds since the epoch (00:00:00 UTC, January 1, 1970)
  protected:
    Timestamp& copyFrom(std::tm value);
    std::tm& copyTo(std::tm& value) const;
  public:
    Timestamp() {}
    Timestamp(const Timestamp& value): ftime(value) {}
    Timestamp(std::time_t value): ftime(value) {}
    Timestamp(const std::tm& value) { copyFrom(value); }
    Timestamp(int year, int month, int day, int hour = 0, int min = 0, int sec = 0);
    operator std::time_t() const { return ftime; }
    operator std::tm() const { std::tm result; return copyTo(result); }
    std::string str() const;
    void decode(int& year, int& month, int& day) const { int dummy; decode(year,month,day,dummy,dummy,dummy); }
    void decode(int& year, int& month, int& day, int& hour, int& min, int& sec) const;
    int dow() const;
    Timestamp& operator=(const Timestamp& value) { ftime = value; return *this; }
    Timestamp& operator=(const std::time_t value) { ftime = value; return *this; }
    Timestamp& operator=(const std::tm& value) { return copyFrom(value); }
    Timestamp& operator++() { ++ftime; return *this; }
    Timestamp operator++(int) { Timestamp result(*this); ++ftime; return result; }
    Timestamp& operator--() { --ftime; return *this; }
    Timestamp operator--(int) { Timestamp result(*this); --ftime; return result; }
    Timestamp& operator+=(int value) { ftime += value; return *this; }
    Timestamp& operator-=(int value) { ftime -= value; return *this; }
    Timestamp operator+(int value) { return Timestamp(*this) += value; }
    Timestamp operator-(int value) { return Timestamp(*this) -= value; }
    int operator-(const Timestamp& value) { return ftime-value; }
    bool operator==(const Timestamp& value) { return ftime == value; }
    bool operator!=(const Timestamp& value) { return ftime != value; }
    bool operator<(const Timestamp& value) { return ftime < value; }
    bool operator<=(const Timestamp& value) { return ftime <= value; }
    bool operator>(const Timestamp& value) { return ftime > value; }
    bool operator>=(const Timestamp& value) { return ftime >= value; }
    std::ostream& put(std::ostream& stream) const { return stream << str(); }
};

inline std::ostream& operator<<(std::ostream& stream, const Timestamp& t) {
  return t.put(stream);
}

Timestamp IBXX_EXPORT now();

} // namespace IB

#endif // __IBXX_CXXBITS_TIMESTAMP_H__
