/*******************************************************************************

Copyright (C) 2004 Andrea Biardi - LOGICOMM

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

*******************************************************************************/

#include <libinterbasexx/error.h>
#include <libinterbasexx/info.h>

namespace IB {

void DatabaseInfo::getInfo(isc_db_handle handle) {

  // wipe out current information

  flevel = fconnection = fodsMajor = fodsMinor = fdialect = fpageSize = fpages = fbuffers = fusedMemory = 0;
  fversion = fhost = fname = "";
  fdesc.clear();
  fusers.clear();

  ISC_STATUS v[20];

  // first call isc_version() to get some description strings (this is done
  // through a static callback function); isc_version() does not take a status
  // vector as most other functions do, so we do no error checking here

  isc_version(&handle,(isc_callback) pushDesc,this);

  // prepare a buffer for the items to be requested

  char items[] = {
    isc_info_base_level,
    isc_info_db_id,
    isc_info_version,
    isc_info_ods_version,
    isc_info_ods_minor_version,
    isc_info_db_SQL_dialect,
    isc_info_page_size,
    isc_info_allocation,
    isc_info_num_buffers,
    isc_info_current_memory,
    isc_info_user_names
  };

  // set up a buffer for the reply; it should be big enough, or information gets
  // truncated; though the doc says isc_info_truncated is put in the last byte
  // of the buffer, you can expect a segmentation fault in this case, so it's of
  // no use testing it anyway..

  char buf[4096];
  if (isc_database_info(v,&handle,sizeof items,items,sizeof buf,buf)) throw EInterbase(v);

  for (char* p = buf; *p != isc_info_end;) {

    // c = item
    // n = cluster length
    // v = cluster data (value)

    char c = *p++;
    short n = short(isc_vax_integer(p,2));
    p += 2;
    char* v = p;
    p += n;

    switch (c) {

      case isc_info_base_level:
        ++v;
        flevel = *v;
        break;

      case isc_info_db_id:
        fconnection = *v++;
        fname.assign(v+1,*v);
        v += *v+1;
        fhost.assign(v+1,*v);
        break;

      case isc_info_version:
        ++v;
        fversion.assign(v+1,*v);
        break;

      case isc_info_ods_version:
        fodsMajor = isc_vax_integer(v,n);
        break;

      case isc_info_ods_minor_version:
        fodsMinor = isc_vax_integer(v,n);
        break;

      // for the dialect, versions of Interbase earlier than 6.0 might return a
      // value of zero, whereas it should really be dialect 1 (thanks Iain Reid
      // <iain@xenocide.demon.co.uk>)

      case isc_info_db_SQL_dialect:
        fdialect = isc_vax_integer(v,n);
        if (!fdialect) fdialect = 1;
        break;

      case isc_info_page_size:
        fpageSize = isc_vax_integer(v,n);
        break;

      case isc_info_allocation:
        fpages = isc_vax_integer(v,n);
        break;

      case isc_info_num_buffers:
        fbuffers = isc_vax_integer(v,n);
        break;

      case isc_info_current_memory:
        fusedMemory = isc_vax_integer(v,n);
        break;

      case isc_info_user_names: {
        std::string s;
        fusers.push_back(s.assign(v+1,*v));
        break;
      }

    }

  }

}

isc_callback DatabaseInfo::pushDesc(DatabaseInfo* n, char* c) {
  n->fdesc.push_back(c);
  return 0;
}

void StatementInfo::getInfo(isc_stmt_handle handle) {

  // the statement must have been already executed; preparing is not enough, in
  // the sense that no useful information will be gathered; for "select"
  // statements, in particular, one could argue that this function is completely
  // useless, since you can simply increment a local variable to keep track of
  // how many records were actually fetched; "update", "delete" and "insert" SQL
  // statements, however, may want retrieve the number of affected rows with
  // this function, but only after execution

  // wipe out current information

  fupdates = fdeletes = fselects = finserts = 0;

  ISC_STATUS v[20];
  char items = isc_info_sql_records;

  // setup a buffer for the reply (32 bytes should be enough, but let's be on
  // the safe side)

  char buf[128];
  if (isc_dsql_sql_info(v,&handle,1,&items,sizeof buf,buf)) throw EInterbase(v);

  // now see if we got a reply

  if (buf[0] == isc_info_sql_records) {

    for (char* p = buf+3; *p != isc_info_end;) {

      // c = item
      // n = cluster length
      // v = cluster data (value)

      char c = *p++;
      short n = short(isc_vax_integer(p,2));
      p += 2;
      int v = isc_vax_integer(p,n);
      p += n;

      switch (c) {

        case isc_info_req_update_count:
          fupdates = v;
          break;

        case isc_info_req_delete_count:
          fdeletes = v;
          break;

        case isc_info_req_select_count:
          fselects = v;
          break;

        case isc_info_req_insert_count:
          finserts = v;
          break;

      }
    }
  }

}

} // namespace IB
