/*******************************************************************************

Copyright (C) 2003 Andrea Biardi - LOGICOMM

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

*******************************************************************************/

#include <cassert>
#include <libinterbasexx/query.h>

namespace IB {

void Query::Out::resize(short value) {

  // resize the XSQLDA buffer so that it can hold up to a given number of SQL
  // variables

  fxsqlda.resize(XSQLDA_LENGTH(value));
  xsqlda()->version = SQLDA_VERSION1;
  xsqlda()->sqln = value;

}

Query::~Query() {

  // avoid throwing exceptions in the destructor; act silently.

  try {

    // it is not clear to me if we need to close() the query, because we'll be
    // freeing the statement in a moment; maybe we could do without (it should
    // be noted that named cursors, which are currently not implemented here,
    // *do* require closing the statement before freeing its handle) -- anyway
    // this doesn't seem to do any harm

    if (active()) close();

  }
  catch (const EInterbase&) {}

}

void Query::prepare() {

  // make sure the query isn't already prepared

  IBXX_ASSERT((!prepared()) || (fstatement.prepared()));
  checkNotPrepared();

  // first, have a prepared statement; note that the statement itself might be
  // already prepared, if during a previous call to Query::prepare() we failed
  // to initialize our own things

  if (!fstatement.prepared()) fstatement.prepare();

  // then, we have to finish up by initializing output things

  ISC_STATUS v[20];
  isc_stmt_handle h = handle();

  // bind the output XSQLDA; note that the call to isc_dsql_describe() must be
  // done once more if the XSQLDA gets resized (we have more output columns than
  // originally allocated)

  if (isc_dsql_describe(v,&h,database().dialect(),fout.xsqlda())) throw EInterbase(v);

  if (fout.xsqlda()->sqld > fout.xsqlda()->sqln) {
    fout.resize(fout.xsqlda()->sqld);
    if (isc_dsql_describe(v,&h,database().dialect(),fout.xsqlda())) throw EInterbase(v);
  }

  // prepare the buffer for output columns; as in Statement::prepare(), we share
  // a common buffer for all SQL variables

  if (!fout.xsqlda()->sqld) throw EInterbase("Query returns no columns (output XSQLDA is empty)");
  fout.prepare();

  // finally, set our own "prepared" state flag

  fstate |= Prepared;

}

void Query::open() {

  // make sure the query isn't already active

  checkNotActive();

  // prepare and execute the statement

  if (!prepared()) prepare();
  fstatement.exec();

  // mark the query active

  fstate |= Active;

}

void Query::close() {

  // make sure the query is active

  checkActive();

  // call isc_dsql_free_statement() as in Statement::~Statement(), but using
  // DSQL_close instead of DSQL_drop, so the statement is available for further
  // execution (maybe with different input parameters)

  ISC_STATUS v[20];
  isc_stmt_handle h = handle();

  if (isc_dsql_free_statement(v,&h,DSQL_close)) throw EInterbase(v);

  // mark the query inactive (note: the statement remains in its "prepared"
  // state)

  fstate &= ~ Active;
  IBXX_ASSERT(h);
  IBXX_ASSERT(h == handle());

}

bool Query::fetch() {

  // make sure the query is active

  checkActive();

  // isc_dsql_fetch() returns 100 when there's no more data to retrieve

  ISC_STATUS v[20];
  isc_stmt_handle h = handle();

  switch (isc_dsql_fetch(v,&h,database().dialect(),fout.xsqlda())) {
    case 0:
      return true;
    case 100:
      return false;
    default:
      throw EInterbase(v);
  }

}

} // namespace IB
