/*******************************************************************************

Copyright (C) 2003 Andrea Biardi - LOGICOMM

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

*******************************************************************************/

#include <algorithm>
#include <cassert>
#include <cstring>
#include <ctime>
#include <libinterbasexx/error.h>
#include <libinterbasexx/sqlinvar.h>

namespace IB {

void SqlInVar::copy(const SqlInVar& src) {

  switch (src.type()) {

    case 0:
      return clear();

    case Numeric:
      return set_n(src.n());

    case Time:
      return set_t(src.t());

    case String:
      return set_s(src.s());

    case Array:
      return set_a(src.a());

    case Blob:
      return set_b(src.b());

    default:
      IBXX_ASSERT(false);

  }

}

const IB::Numeric& SqlInVar::n() const {

  static IB::Numeric null;
  switch (ftype) {

    case 0:
      return null;

    case Numeric:
      return fvalue_n;

    default:
      throw cannotConvert("Input parameter",ftype,Numeric);

  }

}

const Timestamp& SqlInVar::t() const {

  static Timestamp null;
  switch (ftype) {

    case 0:
      return null;

    case Time:
      return fvalue_t;

    default:
      throw cannotConvert("Input parameter",ftype,Time);

  }

}

const std::string& SqlInVar::s() const {

  static std::string null;
  switch (ftype) {

    case 0:
      return null;

    case String:
      return fvalue_s;

    default:
      throw cannotConvert("Input parameter",ftype,String);

  }

}

const IB::Array& SqlInVar::a() const {

  static IB::Array null;
  switch (ftype) {

    case 0:
      return null;

    case Array:
      return fvalue_a;

    default:
      throw cannotConvert("Input parameter",ftype,Array);

  }

}


const IB::Blob& SqlInVar::b() const {

  static IB::Blob null;
  switch (ftype) {

    case 0:
      return null;

    case Blob:
      return fvalue_b;

    default:
      throw cannotConvert("Input parameter",ftype,Blob);

  }

}

void SqlInVar::copyToXSQLVAR(XSQLVAR* v) const {

  // (attempt to) copy current value to the given XSQLVAR*

  if (null()) {

    // here we have a null parameter; if the column does not allow nulls we are
    // in the unfortunate case of not being able to provide much information in
    // the error message

    if (!(v->sqltype & 1)) EInterbase("Input parameter: null value not allowed");

    *v->sqlind = -1;
    return;

  }

  // clear the null indicator and copy the value

  if (v->sqltype & 1) *v->sqlind = 0;

  switch (v->sqltype & ~1) {

    case SQL_SHORT:
      *(short*) v->sqldata = short(scale(n(),v->sqlscale));
      return;

    case SQL_LONG:
      *(long*) v->sqldata = long(scale(n(),v->sqlscale));
      return;

    case SQL_INT64:
      *(ISC_INT64*) v->sqldata = ISC_INT64(scale(n(),v->sqlscale));
      return;

    case SQL_FLOAT:
      *(float*) v->sqldata = float(n());
      return;

    case SQL_DOUBLE:
      *(double*) v->sqldata = double(n());
      return;

    case SQL_TIMESTAMP: {
      std::tm m = t();
      return isc_encode_timestamp(&m,(ISC_TIMESTAMP*) v->sqldata);
    }

    case SQL_TYPE_DATE: {
      std::tm m = t();
      return isc_encode_sql_date(&m,(ISC_DATE*) v->sqldata);
    }

    case SQL_TYPE_TIME: {
      std::tm m = t();
      return isc_encode_sql_time(&m,(ISC_TIME*) v->sqldata);
    }

    case SQL_TEXT:
      std::memset(v->sqldata,' ',v->sqllen);
      s().copy(v->sqldata,v->sqllen);
      return;

    // SQL_VARYING strings are truncated to SHRT_MAX characters; no exception is
    // thrown, however, in conformance to what happens for SQL_TEXT, where
    // Interbase does impose a stricter limit (v->sqllen)

    case SQL_VARYING: {
      const std::string& t = s();
      short n = short(std::min(t.length(),std::string::size_type(SHRT_MAX)));
      *(short*) v->sqldata = n;
      t.copy(v->sqldata+sizeof(short),n);
      return;
    }

    case SQL_ARRAY:
      *(ISC_QUAD*) v->sqldata = a().id();
      return;

    case SQL_BLOB:
      *(ISC_QUAD*) v->sqldata = b().id();
      return;

    default:
      throw unknownSqlType("Input parameter",v->sqltype);

  }

}

} // namespace IB
