/*******************************************************************************

Copyright (C) 2003 Andrea Biardi - LOGICOMM

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

*******************************************************************************/

#ifndef __IBXX_SQLOUTVAR_H__
#define __IBXX_SQLOUTVAR_H__

/*******************************************************************************

Encapsulation of SQL variables for output columns.

Note: though this class does not make the copy constructor private, it should
be noted that the pointer to the SQL variable, not the actual value, is copied.

*******************************************************************************/

#include <cassert>
#include <ctime>
#include <iostream>
#include <libinterbasexx/array.h>
#include <libinterbasexx/blob.h>
#include <libinterbasexx/config.h>
#include <libinterbasexx/cxxbits/timestamp.h>
#include <libinterbasexx/numeric.h>
#include <libinterbasexx/sqlvar.h>
#include <string>
#include <ibase.h>

namespace IB {

class IBXX_EXPORT SqlOutVar: public SqlVar {
  private:
    const XSQLVAR* fvar; // (read-only) pointer to SQL variable
  private:
    SqlOutVar& operator=(const SqlOutVar&) { return *this; }
  public:
    SqlOutVar(const XSQLVAR* var): fvar(var) {}
    int type() const;
    short sqltype() const { return fvar->sqltype; }
    short sqlsubtype() const { return fvar->sqlsubtype; }
    short sqlscale() const { return fvar->sqlscale; }
    short sqllen() const { return fvar->sqllen; }
    const char* sqldata() const { return fvar->sqldata; }
    std::string desc() const;
    std::string name() const;
    std::string alias() const;
    std::string table() const;
    std::string owner() const;
    bool null() const { return (fvar->sqltype & 1) && (*fvar->sqlind == -1); }
    template<class T> T n() const;
    template<class T> T t() const;
    template<class T> T s() const;
    template<class T> T a() const;
    template<class T> T b() const;
    int i() const { return n<int>(); }
    double f() const { return n<double>(); }
    IB::Numeric n() const { return n<IB::Numeric>(); }
    Timestamp t() const { return t<Timestamp>(); }
    std::string s() const { return s<std::string>(); }
    IB::Array a() const { return a<IB::Array>(); }
    IB::Blob b() const { return b<IB::Blob>(); }
    std::ostream& put(std::ostream& stream) const { return SqlVar::put(stream,*this); }
};

inline int SqlOutVar::type() const {

  switch (sqltype() & ~1) {

    case SQL_SHORT:
    case SQL_LONG:
    case SQL_INT64:
    case SQL_FLOAT:
    case SQL_DOUBLE:
      return Numeric;

    case SQL_TIMESTAMP:
    case SQL_TYPE_DATE:
    case SQL_TYPE_TIME:
      return Time;

    case SQL_TEXT:
    case SQL_VARYING:
      return String;

    case SQL_ARRAY:
      return Array;

    case SQL_BLOB:
      return Blob;

    default:
      throw unknownSqlType(desc(),sqltype());

  }

}

template<class T> T SqlOutVar::n() const {

  // the T class should provide the following constructors:
  //   - T()
  //   - T(short)
  //   - T(long)
  //   - T(ISC_INT64)
  //   - T(float)
  //   - T(double)
  //   - T(const T&)
  //
  // it should also provide the /= operator used in unscale(), since SQL_SHORT,
  // SQL_LONG and SQL_INT64 might be scaled floating point values

  switch (sqltype() & ~1) {

    case SQL_SHORT: {
      if (null()) return T();
      T result(T(*(const short*) sqldata()));
      return unscale(result,sqlscale());
    }

    case SQL_LONG: {
      if (null()) return T();
      T result(T(*(const long*) sqldata()));
      return unscale(result,sqlscale());
    }

    case SQL_INT64: {
      if (null()) return T();
      T result(T(*(const ISC_INT64*) sqldata()));
      return unscale(result,sqlscale());
    }

    case SQL_FLOAT:
      return null() ? T() : T(*(const float*) sqldata());

    case SQL_DOUBLE:
      return null() ? T() : T(*(const double*) sqldata());

    default:
      throw cannotConvert(desc(),type(),Numeric);

  }

}

template<class T> T SqlOutVar::t() const {

  // the T class should provide the following constructors:
  //   - T()
  //   - T(const std::tm&)
  //   - T(const T&)

  switch (sqltype() & ~1) {

    case SQL_TIMESTAMP: {
      if (null()) return T();
      std::tm result;
      isc_decode_timestamp((ISC_TIMESTAMP*) sqldata(),&result);
      return T(result);
    }

    case SQL_TYPE_DATE: {
      if (null()) return T();
      std::tm result;
      isc_decode_sql_date((ISC_DATE*) sqldata(),&result);
      IBXX_ASSERT(!result.tm_hour);
      IBXX_ASSERT(!result.tm_min);
      IBXX_ASSERT(!result.tm_sec);
      IBXX_ASSERT(!result.tm_isdst);
      return T(result);
    }

    case SQL_TYPE_TIME: {
      if (null()) return T();
      std::tm result;
      isc_decode_sql_time((ISC_TIME*) sqldata(),&result);
      IBXX_ASSERT(!result.tm_year);
      IBXX_ASSERT(!result.tm_mon);
      IBXX_ASSERT(!result.tm_mday);
      IBXX_ASSERT(!result.tm_wday);
      IBXX_ASSERT(!result.tm_yday);
      return T(result);
    }

    default:
      throw cannotConvert(desc(),type(),Time);

  }

}

template<class T> T SqlOutVar::s() const {

  // the T class should provide the following constructors:
  //   - T()
  //   - T(const std::string&)
  //   - T(const T&)

  switch (sqltype() & ~1) {

    case SQL_TEXT: {
      if (null()) return T();
      std::string result;
      return T(result.assign(sqldata(),sqllen()));
    }

    case SQL_VARYING: {
      if (null()) return T();
      std::string result;
      return T(result.assign(sqldata()+sizeof(short),*(short*) sqldata()));
    }

    default:
      throw cannotConvert(desc(),type(),String);

  }

}

template<class T> T SqlOutVar::a() const {

  // the T class should provide the following constructors:
  //   - T()
  //   - T(const ISC_QUAD&)
  //   - T(const T&)

  switch (sqltype() & ~1) {

    case SQL_ARRAY:
      return null() ? T() : T(*(const ISC_QUAD*) sqldata());

    default:
      throw cannotConvert(desc(),type(),Array);

  }

}

template<class T> T SqlOutVar::b() const {

  // the T class should provide the following constructors:
  //   - T()
  //   - T(const ISC_QUAD&)
  //   - T(const T&)

  switch (sqltype() & ~1) {

    case SQL_BLOB:
      return null() ? T() : T(*(const ISC_QUAD*) sqldata());

    default:
      throw cannotConvert(desc(),type(),Blob);

  }

}

inline std::ostream& operator<<(std::ostream& stream, const SqlOutVar& v) {
  return v.put(stream);
}

} // namespace IB

#endif // __IBXX_SQLOUTVAR_H__
