/*******************************************************************************

Copyright (C) 2003 Andrea Biardi - LOGICOMM

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

*******************************************************************************/

#ifndef __IBXX_SQLVAR_H__
#define __IBXX_SQLVAR_H__

/*******************************************************************************

Encapsulation of SQL variables (not a "real" class: it just contains code
common to both input parameters and output columns).

*******************************************************************************/

#include <iostream>
#include <libinterbasexx/config.h>
#include <libinterbasexx/cxxbits/format.h>
#include <libinterbasexx/error.h>
#include <string>

namespace IB {

class IBXX_EXPORT SqlVar {
  public:
    enum Types {
      Numeric = 0x0001, // numeric variable (integer or floating point)
      Time    = 0x0010, // time (date, time or date & time)
      String  = 0x0100, // string (fixed or variable size)
      Array   = 0x1000, // array ID
      Blob    = 0x2000  // blob ID
    };
  protected:
    static EInterbase unknownSqlType(const std::string& desc, int sqltype) { return EInterbase(format("%s: Unknown SQL data type %d",desc,sqltype)); }
    static EInterbase cannotConvert(const std::string& desc, int fromType, int toType) { return cannotConvert(desc,typeDesc(fromType),typeDesc(toType)); }
    static EInterbase cannotConvert(const std::string& desc, const std::string& fromType, const std::string& toType) { return EInterbase(format("%s: Cannot convert %s to %s",desc,fromType,toType)); }
    template<class T> T scale(const T& value, int scale) const;
    template<class T> T& unscale(T& value, int scale) const;
  public:
    static std::string typeDesc(int type);
    template<class T> std::ostream& put(std::ostream& stream, const T& value) const;
};

template<class T> T SqlVar::scale(const T& value, int scale) const {

  // scaled numbers are defined as having a negative scale; for example, with an
  // input parameter whose scale is -2, we need to supply 12345 if we want to
  // get the value 123.45
  //
  // see also unscale(), which does the opposite thing; please note that scale()
  // and unscale() have slightly different prototypes

  T result = value;
  while (scale++ < 0) result *= 10;

  return result;

}

template<class T> T& SqlVar::unscale(T& value, int scale) const {

  // scaled numbers are defined as having a negative scale; for example, the
  // value 12345 from an ouput column with a scale of -2 is actually 123.45
  //
  // see also scale(), which does the opposite thing; please note that scale()
  // and unscale() have slightly different prototypes

  while (scale++ < 0) value /= 10;
  return value;

}

template<class T> std::ostream& SqlVar::put(std::ostream& stream, const T& value) const {

  if (value.null()) return stream << "<null>";

  switch (value.type()) {

    case Numeric:
      return stream << value.n();

    case Time:
      return stream << value.t();

    case String:
      return stream << value.s();

    case Array:
      return stream << value.a();

    case Blob:
      return stream << value.b();

  }

  // flow does not reach this point, as unknown data types are expected to throw
  // an exception in type(); but let's make compilers happy by returning
  // something

  return stream;

}

} // namespace IB

#endif // __IBXX_SQLVAR_H__
