/*******************************************************************************

Copyright (C) 2003 Andrea Biardi - LOGICOMM

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

*******************************************************************************/

#include <cassert>
#include <libinterbasexx/statement.h>
#include <libinterbasexx/xsqlda.h>

namespace IB {

void Statement::In::resize(short value) {

  // resize the master copy of our input parameters

  fparams.resize(value);

  // resize the XSQLDA buffer so that it can hold up to a given number of SQL
  // variables

  fxsqlda.resize(XSQLDA_LENGTH(value));
  xsqlda()->version = SQLDA_VERSION1;
  xsqlda()->sqln = value;

}

void Statement::In::prepare() {

  // sanity check

  IBXX_ASSERT(xsqlda()->sqln == count());
  IBXX_ASSERT(xsqlda()->sqld <= xsqlda()->sqln);

  // prepare the buffer for the actual values of SQL variables

  prepareXSQLVarsBuffer(xsqlda(),fvars);

}

void Statement::In::push() {

  // sanity check

  IBXX_ASSERT(xsqlda()->sqln == count());
  IBXX_ASSERT(xsqlda()->sqld <= count());

  // copy at most xsqlda()->sqld variables from fparams to the input XSQLDA
  // buffer

  XSQLVAR* v = xsqlda()->sqlvar;
  for (int i = 0; i < xsqlda()->sqld; ++i, ++v) fparams[i].copyToXSQLVAR(v);

}

Statement::Statement(const Database& database, const Transaction& transaction, const std::string& text): fdatabase(database), ftransaction(transaction), ftext(text), fhandle(0), fprepared(false), fin(countQuestionMarks(text)) {

  // allocate the statement

  ISC_STATUS v[20];
  isc_db_handle h = fdatabase.handle();

  if (isc_dsql_allocate_statement(v,&h,&fhandle)) throw EInterbase(v);

  // we should now have a handle

  IBXX_ASSERT(fhandle);

}

Statement::~Statement() {

  // avoid throwing exceptions in the destructor; act silently (no error
  // checking here).

  ISC_STATUS v[20];
  isc_dsql_free_statement(v,&fhandle,DSQL_drop);

}

void Statement::prepare() {

  // make sure the statement isn't already prepared

  checkNotPrepared();

  // setup a status vector and make a local copy of the transaction handle

  ISC_STATUS v[20];
  isc_tr_handle h = ftransaction.handle();

  // prepare the statement -- isc_dsql_prepare() will do

  if (isc_dsql_prepare(v,&h,&fhandle,0,(char*) ftext.c_str(),fdatabase.dialect(),0)) throw EInterbase(v);

  // bind the input XSQLDA; we should proceed like this:
  //   - call isc_dsql_describe_bind(), which - among other things - computes
  //     fin.xsqlda()->sqld
  //   - check if fin.xsqlda()->sqld is greater than fin.xsqlda()->sqln; the
  //     latter is always equal to fin.count()
  //   - if so, resize the XSQLDA and start over
  //
  // however, we already know the number of input parameters, as setText()
  // counts the number of question marks in the SQL string, and we already set
  // fin.xsqlda()->sqln to this number; thus, we can be pretty sure that our
  // input XSQLDA is already ok

  if (isc_dsql_describe_bind(v,&fhandle,fdatabase.dialect(),fin.xsqlda())) throw EInterbase(v);

  // prepare the buffer for input parameters; instead of having a single
  // allocation for each parameter, we use a common buffer (this call resizes
  // the buffer and initializes pointers for each SQL variable)

  fin.prepare();

  // the statement is now ready to execute (multiple times); remember this
  // somewhere

  fprepared = true;

}

void Statement::setText(const std::string& value) {

  // unprepare the statement

  fprepared = false;

  // update the text; note that we are counting the number of question marks in
  // the SQL text so that we can resize In automatically; this method is not
  // reliable enough, as we might be allocating more input parameters than we
  // actually need; take, for example:
  //
  //   insert into table1 (column1) values ("??")
  //
  // in this case a literal string should be inserted, but we will allocate
  // space for two input parameters anyway; and:
  //
  //   delete ?? from ??
  //
  // ends up with four input parameters, though this statement won't even
  // execute; anyway, it's just some bytes wasted, and most importantly, we're
  // always on the safe side
  //
  // if you need to know the exact number of input parameters, the only thing
  // you can do is:
  //   - prepare() the statement manually
  //   - access in().xsqlda()->sqld

  fin.resize(countQuestionMarks(ftext = value));

}

void Statement::exec() {

  // if needed, prepare the statement

  if (!fprepared) prepare();

  // setup a status vector and make a local copy of the transaction handle

  ISC_STATUS v[20];
  isc_tr_handle h = ftransaction.handle();

  // copy input parameters to SQL variables

  fin.push();

  // execute the statement

  if (isc_dsql_execute(v,&h,&fhandle,fdatabase.dialect(),fin.xsqlda())) throw EInterbase(v);

}

} // namespace IB
