/*******************************************************************************

Copyright (C) 2003 Andrea Biardi - LOGICOMM

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

*******************************************************************************/

#ifndef __IBXX_STATEMENT_H__
#define __IBXX_STATEMENT_H__

/*******************************************************************************

Encapsulation of Interbase C API for statements.

*******************************************************************************/

#include <algorithm>
#include <libinterbasexx/config.h>
#include <libinterbasexx/cxxbits/cbuffer.h>
#include <libinterbasexx/database.h>
#include <libinterbasexx/error.h>
#include <libinterbasexx/sqlinvar.h>
#include <libinterbasexx/transaction.h>
#include <string>
#include <vector>
#include <ibase.h>

namespace IB {

class IBXX_EXPORT Statement {
  public:
    class In {
      friend class Statement;
      private:
        CBuffer fxsqlda;               // input XSQLDA
        CBuffer fvars;                 // input buffer for SQL variables
        std::vector<SqlInVar> fparams; // input parameters (master copy)
      private:
        In(const In&) {}
        In& operator=(const In&) { return *this; }
      protected:
        void resize(short value);
        void prepare();
        void push();
      public:
        In(short count = 0) { resize(count); }
        const XSQLDA* xsqlda() const { return (const XSQLDA*) fxsqlda.data(); }
        XSQLDA* xsqlda() { return (XSQLDA*) fxsqlda.data(); }
        short count() const { return short(fparams.size()); }
        void clear() { for (std::vector<SqlInVar>::iterator p = fparams.begin(); p != fparams.end(); ++p) p->clear(); }
        const SqlInVar& operator[](short index) const { return fparams[index]; }
        SqlInVar& operator[](short index) { return fparams[index]; }
    };
  private:
    const Database& fdatabase;       // database
    const Transaction& ftransaction; // transaction
    std::string ftext;               // SQL text
    isc_stmt_handle fhandle;         // statement handle
    bool fprepared;                  // true when statement is prepared
    In fin;                          // statement input
  private:
    Statement(const Statement& src): fdatabase(src.database()), ftransaction(src.transaction()) {}
    Statement& operator=(const Statement&) { return *this; }
  protected:
    static short countQuestionMarks(const std::string& s) { return short(std::min(std::count(s.begin(),s.end(),'?'),SHRT_MAX)); }
    void checkPrepared() const { if (!fprepared) throw EInterbase("Statement is not prepared"); }
    void checkNotPrepared() const { if (fprepared) throw EInterbase("Statement is already prepared"); }
  public:
    Statement(const Database& database, const Transaction& transaction, const std::string& text = "");
    ~Statement();
    const Database& database() const { return fdatabase; }
    const Transaction& transaction() const { return ftransaction; }
    const std::string& text() const { return ftext; }
    void setText(const std::string& value);
    isc_stmt_handle handle() const { return fhandle; }
    bool prepared() const { return fprepared; }
    void prepare();
    void exec();
    const In& in() const { return fin; }
    In& in() { return fin; }
};

inline void execute(const Database& database, const Transaction& transaction, const std::string& text, unsigned short dialect = 0) {

  // this function (in two variants, see below) is a simple replacement of the
  // Statement class, especially useful for short SQL statements; limitations
  // are as follows:
  //   - no input parameters (good if you know you don't need any)
  //   - no preparation (good if the SQL statement is to be executed just once,
  //     bad otherwise)
  //   - no post-execution information (StatementInfo class not applicable)
  //
  // this first variant is useful to execute statements against a connected
  // database, in the context of an active transaction; if you need to issue a
  // "create database" SQL statement, then you don't have a database or a
  // transaction yet and thus use the second variant below
  //
  // dialect defaults to 0, in which case we take the one of the given database

  isc_db_handle dh = database.handle();
  isc_tr_handle th = transaction.handle();

  if (!dialect) dialect = database.dialect();

  // call isc_dsql_execute_immediate to do the job; note that we don't pass an
  // input XSQLDA: if you do need input parameters, use the Statement class
  // instead

  ISC_STATUS v[20];
  if (isc_dsql_execute_immediate(v,&dh,&th,0,(char*) text.c_str(),dialect,0)) throw EInterbase(v);

}

inline void execute(const std::string& text, unsigned short dialect = 0) {

  // same as above, useful to execute SQL statements outside of a transaction
  // context

  isc_db_handle dh = 0;
  isc_tr_handle th = 0;

  if (!dialect) dialect = IBXX_DEFAULT_DIALECT;

  // call isc_dsql_execute_immediate to do the job; note that we don't pass an
  // input XSQLDA here

  ISC_STATUS v[20];
  if (isc_dsql_execute_immediate(v,&dh,&th,0,(char*) text.c_str(),dialect,0)) throw EInterbase(v);

}

} // namespace IB

#endif // __IBXX_STATEMENT_H__
