/*******************************************************************************

Copyright (C) 2003 Andrea Biardi - LOGICOMM

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

*******************************************************************************/

#include <cassert>
#include <libinterbasexx/cxxbits/format.h>
#include <libinterbasexx/cxxbits/strmisc.h>
#include <libinterbasexx/parameterblock.h>
#include <libinterbasexx/transaction.h>

namespace IB {

Transaction::~Transaction() {

  // avoid throwing exceptions in the destructor; act silently.

  try {
    if (active()) rollback();
  }
  catch (const EInterbase&) {}

}

void Transaction::reserve(const std::string& table, int flags) {

  // make sure the transaction isn't already active

  checkNotActive();

  // before playing with the map, work around a bug in Interbase 6, that
  // requires table names to be all uppercase here
  //
  // note that invalid names will prevent the transaction to work, throwing an
  // exception by that time start() is called
  //
  // also note that each time a table is reserved, any previous reservation for
  // the same table is automatically discarded

  freservations[ucase_c(table)] = flags;

}

void Transaction::unreserve(const std::string& table) {

  // make sure the transaction isn't already active

  checkNotActive();

  // check if the reservation is there (name must be converted to uppercase or
  // it won't work)

  Reservations::iterator p;
  if ((p = freservations.find(ucase_c(table))) == freservations.end()) throw EInterbase(format("Table %s has not been reserved for this transaction",table));

  freservations.erase(p);

}

void Transaction::start() {

  // make sure the transaction isn't already active

  checkNotActive();

  // set up a TPB (transaction parameter block)
  //
  // the following values are not passed, because are implicitely set by the
  // Interbase API:
  //   - ReadWrite
  //   - Concurrency
  //   - Wait

  ParameterBlock tpb;
  tpb.push(isc_tpb_version3);

  if (fflags & ReadOnly) tpb.push(isc_tpb_read);

  switch (fflags & IsolationMask) {

    case ReadCommitted:
      tpb.push(isc_tpb_read_committed);
      break;

    case Versioning:
      tpb.push(isc_tpb_read_committed);
      tpb.push(isc_tpb_rec_version);
      break;

    case Consistency:
      tpb.push(isc_tpb_consistency);
      break;

  }

  if (fflags & NoWait) tpb.push(isc_tpb_nowait);

  // append table reservation entries (if any)

  for (Reservations::const_iterator p = freservations.begin(); p != freservations.end(); ++p) {

    // the Interbase API documentation explains the table reservation
    // functionality in a very approximate way:
    //   - it doesn't state the format of these fields in the TPB (or, rather,
    //     it does it in a misleading way)
    //   - it doesn't say whether isc_tpb_shared is really to be included in the
    //     TPB, even though it is the default option
    //   - it mentions isc_tpb_exclusive, but doesn't clarify its use
    //
    // moreover, the code contains a bug that requires table names to be
    // uppercase..

    switch (p->second & ShareMask) {

      case Protected:
        tpb.push(isc_tpb_protected);
        break;

      case Exclusive:
        tpb.push(isc_tpb_exclusive);
        break;

      // not sure whether isc_tpb_shared is implicitely set by the Interbase
      // API; let's include this anyway, just to be on the safe side..

      default:
        tpb.push(isc_tpb_shared);
        break;

    }

    tpb.push(char(p->second & LockRead ? isc_tpb_lock_read : isc_tpb_lock_write),p->first);

  }

  // call isc_start_transaction() to do the job

  ISC_STATUS v[20];
  isc_db_handle h = fdatabase.handle();

  if (isc_start_transaction(v,&fhandle,1,&h,tpb.size(),tpb.data())) throw EInterbase(v);

  // we should now have a handle

  IBXX_ASSERT(fhandle);

}

void Transaction::commit(bool retaining) {

  // make sure the transaction is active

  checkActive();

  // decide what to do next

  ISC_STATUS v[20];
  if (retaining) {

    // after the call to isc_commit_retaining(), we should have a non-null
    // handle

    if (isc_commit_retaining(v,&fhandle)) throw EInterbase(v);
    IBXX_ASSERT(fhandle);

  } else {

    // after the call to isc_commit_transaction(), we should have a null handle

    if (isc_commit_transaction(v,&fhandle)) throw EInterbase(v);
    IBXX_ASSERT(!fhandle);

  }

}

void Transaction::rollback(bool retaining) {

  // for read-only transactions, the official documentation suggests using
  // isc_commit_transaction() rather than isc_rollback_transaction(), because
  // "the overhead required to start subsequent transactions is greatly reduced"
  // ("Interbase 6 API Guide", page 75); however, you might still like to call
  // rollback() in your applications

  if (fflags & ReadOnly) return commit(retaining);

  // make sure the transaction is active

  checkActive();

  // decide what to do next

  ISC_STATUS v[20];
  if (retaining) {

    // after the call to isc_rollback_retaining(), we should have a non-null
    // handle

    if (isc_rollback_retaining(v,&fhandle)) throw EInterbase(v);
    IBXX_ASSERT(fhandle);

  } else {

    // after the call to isc_rollback_transaction(), we should have a null
    // handle

    if (isc_rollback_transaction(v,&fhandle)) throw EInterbase(v);
    IBXX_ASSERT(!fhandle);

  }

}

} // namespace IB
