/*******************************************************************************

Copyright (C) 2003 Andrea Biardi - LOGICOMM

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

*******************************************************************************/

#ifndef __IBXX_TRANSACTION_H__
#define __IBXX_TRANSACTION_H__

/*******************************************************************************

Encapsulation of Interbase C API for transaction handling.

*******************************************************************************/

#include <libinterbasexx/config.h>
#include <libinterbasexx/database.h>
#include <libinterbasexx/error.h>
#include <map>
#include <string>
#include <ibase.h>

namespace IB {

class IBXX_EXPORT Transaction {
  public:
    enum Flags {
      ReadWrite          = 0x0001, // read-write transaction (default)
      ReadOnly           = 0x0002, // read-only transaction
      AccessMask         = ReadWrite | ReadOnly,
      Concurrency        = 0x0010, // concurrency (default)
      ReadCommitted      = 0x0020, // read-committed
      Versioning         = 0x0040, // read-committed with version control
      Consistency        = 0x0080, // consistency (use with care)
      IsolationMask      = Concurrency | ReadCommitted | Versioning | Consistency,
      Wait               = 0x0100, // wait until locked resources are released, then retry (default)
      NoWait             = 0x0200, // do not wait: return error lock conflict error
      LockResolutionMask = Wait | NoWait
    };
    enum ReservationFlags {
      LockWrite       = 0x0001, // reserve table for read-write access (default)
      LockRead        = 0x0002, // reserve table for read-only access
      LockMask        = LockWrite | LockRead,
      Shared          = 0x0010, // allow concurrent, shared access (default)
      Protected       = 0x0020, // allow concurrent, restricted access
      Exclusive       = 0x0040, // grant exclusive access
      ShareMask       = Shared | Protected | Exclusive
    };
  protected:
    typedef std::map<std::string,int> Reservations;
  private:
    const Database& fdatabase;  // database
    int fflags;                 // flags, see above
    Reservations freservations; // table reservation entries
    isc_tr_handle fhandle;      // transaction handle
  private:
    Transaction(const Transaction& src): fdatabase(src.database()) {}
    Transaction& operator=(const Transaction&) { return *this; }
  protected:
    void checkActive() const { if (!active()) throw EInterbase("Transaction is not active"); }
    void checkNotActive() const { if (active()) throw EInterbase("Transaction is already active"); }
  public:
    Transaction(const Database& database, int flags = ReadWrite | Concurrency | Wait): fdatabase(database), fflags(flags), fhandle(0) {}
    ~Transaction();
    const Database& database() const { return fdatabase; }
    int flags() const { return fflags; }
    void setFlags(int value) { checkNotActive(); fflags = value; }
    isc_tr_handle handle() const { checkActive(); return fhandle; }
    bool active() const { return fhandle; }
    void reserve(const std::string& table, int flags = ReadWrite | Shared);
    void unreserve(const std::string& table);
    void start();
    void commit(bool retaining = false);
    void rollback(bool retaining = false);
};

} // namespace IB

#endif // __IBXX_TRANSACTION_H__
