/*******************************************************************************

Copyright (C) 2003 Andrea Biardi - LOGICOMM

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE. See the GNU General Public License for more details.

*******************************************************************************/

#include <libinterbasexx/cxxbits/format.h>
#include <libinterbasexx/error.h>
#include <libinterbasexx/xsqlda.h>

namespace IB {

inline int sqlindSize(const XSQLVAR* v) {

  // return the number of bytes needed to store the null indicator for this SQL
  // type (or just 0 if it can never be null)

  return (v->sqltype & 1) ? sizeof(short) : 0;

}

inline int sqldataSize(const XSQLVAR* v) {

  // return the number of bytes needed to store the actual data for this SQL
  // type

  switch (v->sqltype & ~1) {

    case SQL_SHORT:
      return sizeof(short);

    case SQL_LONG:
      return sizeof(long);

    case SQL_INT64:
      return sizeof(ISC_INT64);

    case SQL_FLOAT:
      return sizeof(float);

    case SQL_DOUBLE:
      return sizeof(double);

    case SQL_TIMESTAMP:
      return sizeof(ISC_TIMESTAMP);

    case SQL_TYPE_DATE:
      return sizeof(ISC_DATE);

    case SQL_TYPE_TIME:
      return sizeof(ISC_TIME);

    case SQL_TEXT:
      return v->sqllen;

    case SQL_VARYING:
      return sizeof(short)+v->sqllen;

    case SQL_ARRAY:
    case SQL_BLOB:
      return sizeof(ISC_QUAD);

  }

  throw EInterbase(format("XSQLDA: Unknown SQL data type %d",v->sqltype));

}

void prepareXSQLVarsBuffer(XSQLDA* xsqlda, CBuffer& vars, bool allowResize) {

  // initialize SQL variables to point to the given CBuffer, which is also grown
  // accordingly, if needed (and allowed)

  int i;      // variable index
  int n;      // bytes of buffer used so far
  XSQLVAR* v; // current variable
  char* p;    // next available address in buffer

  // walk through variables and setup both p->sqlind and p->sqldata; note that
  // pointers might not be valid if the buffer is not large enough
  //
  // even so, we still want to complete the loop, so that we know the exact
  // buffer requirements

  for (i = 0, n = 0, v = xsqlda->sqlvar, p = (char*) vars.data(); i < xsqlda->sqld; ++i, ++v) {

    // setup (aligned) pointer for the null indicator (might not be needed)

    int j = sqlindSize(v);
    if ((j = IBXX_ALIGN_SIZE(j)) != 0) {
      v->sqlind = (short*) p;
      p += j;
      n += j;
    }

    // setup (aligned) pointer for data

    j = sqldataSize(v);
    j = IBXX_ALIGN_SIZE(j);
    v->sqldata = p;
    p += j;
    n += j;

  }

  // did we walk outside the allocated buffer? if so, we need to grow it and
  // start over (note that infinite recursion is prevented by disallowing a
  // further resize)

  if (n > vars.size()) {
    if (!allowResize) throw EInterbase(format("XSQLDA: Buffer too small for SQL variables (allocated = %d, required = %d)",vars.size(),n));
    vars.resize(n);
    prepareXSQLVarsBuffer(xsqlda,vars,false);
  }

}

} // namespace IB
